# SPDX-License-Identifier: Apache-2.0
# Copyright © 2022-2024 Intel Corporation

"""Type definitions for cargo manifest files."""

from __future__ import annotations

import collections
import dataclasses
import os
import typing as T


from . import version
from ..mesonlib import MesonException, lazy_property, Version
from .. import mlog

if T.TYPE_CHECKING:
    from typing_extensions import Protocol, Self

    from . import raw
    from .raw import EDITION, CRATE_TYPE, LINT_LEVEL
    from ..wrap.wrap import PackageDefinition

    # Copied from typeshed. Blarg that they don't expose this
    class DataclassInstance(Protocol):
        __dataclass_fields__: T.ClassVar[dict[str, dataclasses.Field[T.Any]]]

_DI = T.TypeVar('_DI', bound='DataclassInstance')

_EXTRA_KEYS_WARNING = (
    "This may (unlikely) be an error in the cargo manifest, or may be a missing "
    "implementation in Meson. If this issue can be reproduced with the latest "
    "version of Meson, please help us by opening an issue at "
    "https://github.com/mesonbuild/meson/issues. Please include the crate and "
    "version that is generating this warning if possible."
)


def fixup_meson_varname(name: str) -> str:
    """Fixup a meson variable name

    :param name: The name to fix
    :return: the fixed name
    """
    return name.replace('-', '_')


class DefaultValue:
    """Base class to converts a raw value from cargo manifest to a meson value

    It returns the value from current manifest, or fallback to the
    workspace value. If both are None, its default value is used. Subclasses can
    override the convert() method to implement custom conversion logic.
    """

    def __init__(self, default: object = None) -> None:
        self.default = default

    def convert(self, v: T.Any, ws_v: T.Any) -> object:
        return v if v is not None else ws_v


class MergeValue(DefaultValue):
    def __init__(self, func: T.Callable[[T.Any, T.Any], object], default: object = None) -> None:
        super().__init__(default)
        self.func = func

    def convert(self, v: T.Any, ws_v: T.Any) -> object:
        return self.func(v, ws_v)


class ConvertValue(DefaultValue):
    def __init__(self, func: T.Callable[[T.Any], object], default: object = None) -> None:
        super().__init__(default)
        self.func = func

    def convert(self, v: T.Any, ws_v: T.Any) -> object:
        return self.func(v if v is not None else ws_v)


class DictMergeValue(ConvertValue):
    """Merge the incoming array of tables with a dictionary;
       a user-provided function maps each table to one of the
       entries of the dictionary."""

    def __init__(self, func: T.Callable[[T.Any], T.List[object]],
                 merge_key: T.Callable[[T.Any], str],
                 out_key: T.Callable[[T.Any], str],
                 base: T.Mapping[str, object] = None) -> None:
        super().__init__(func, base)
        self.merge_key = merge_key
        self.out_key = out_key

    def convert(self, v: T.Any, ws_v: T.Any) -> object:
        out = self.func(v if v is not None else ws_v)
        assert isinstance(out, list) # for mypy
        assert isinstance(self.default, dict) # for mypy

        explicit: T.Set[str] = set(self.merge_key(x) for x in out)
        out_d: T.Dict[str, object] = {self.out_key(x): x for x in out}
        for k, v in self.default.items():
            if self.merge_key(v) not in explicit:
                out_d[self.out_key(v)] = v
        return out_d


def _raw_to_dataclass(raw: T.Mapping[str, object], cls: T.Type[_DI], msg: str,
                      raw_from_workspace: T.Optional[T.Mapping[str, object]] = None,
                      ignored_fields: T.Optional[T.List[str]] = None,
                      **kwargs: DefaultValue) -> _DI:
    """Fixup raw cargo mappings to a dataclass.

    * Inherit values from the workspace.
    * Replaces any `-` with `_` in the keys.
    * Optionally pass values through the functions in kwargs, in order to do
      recursive conversions.
    * Remove and warn on keys that are coming from cargo, but are unknown to
      our representations.

    This is intended to give users the possibility of things proceeding when a
    new key is added to Cargo.toml that we don't yet handle, but to still warn
    them that things might not work.

    :param raw: The raw data to look at
    :param cls: The Dataclass derived type that will be created
    :param msg: the header for the error message. Usually something like "In N structure".
    :param raw_from_workspace: If inheriting from a workspace, the raw data from the workspace.
    :param kwargs: DefaultValue instances to convert values.
    :return: A @cls instance.
    """
    new_dict = {}
    unexpected = set()
    fields = {x.name for x in dataclasses.fields(cls)}
    raw_from_workspace = raw_from_workspace or {}
    ignored_fields = ignored_fields or []
    inherit = raw.get('workspace', False)

    for orig_k, v in raw.items():
        if orig_k == 'workspace':
            continue
        ws_v = None
        if isinstance(v, dict) and v.get('workspace', False):
            # foo.workspace = true, take value from workspace.
            ws_v = raw_from_workspace[orig_k]
            v = None
        elif inherit:
            # foo = {}, give the workspace value, if any, to the converter
            # function in the case it wants to merge values.
            ws_v = raw_from_workspace.get(orig_k)
        k = fixup_meson_varname(orig_k)
        if k not in fields:
            if orig_k not in ignored_fields:
                unexpected.add(orig_k)
            continue
        if k in kwargs:
            new_dict[k] = kwargs[k].convert(v, ws_v)
        else:
            new_dict[k] = v if v is not None else ws_v

    if inherit:
        # Inherit any keys from the workspace that we don't have yet.
        for orig_k, ws_v in raw_from_workspace.items():
            k = fixup_meson_varname(orig_k)
            if k not in fields:
                if orig_k not in ignored_fields:
                    unexpected.add(orig_k)
                continue
            if k in new_dict:
                continue
            if k in kwargs:
                new_dict[k] = kwargs[k].convert(None, ws_v)
            else:
                new_dict[k] = ws_v

    # Finally, set default values.
    for k, convertor in kwargs.items():
        if k not in new_dict and convertor.default is not None:
            new_dict[k] = convertor.default

    if unexpected:
        mlog.warning(msg, 'has unexpected keys', '"{}".'.format(', '.join(sorted(unexpected))),
                     _EXTRA_KEYS_WARNING)

    return cls(**new_dict)


@dataclasses.dataclass
class Package:

    """Representation of a Cargo Package entry, with defaults filled in."""

    name: str
    version: str = "0"
    description: T.Optional[str] = None
    resolver: T.Optional[str] = None
    authors: T.List[str] = dataclasses.field(default_factory=list)
    edition: EDITION = '2015'
    rust_version: T.Optional[str] = None
    documentation: T.Optional[str] = None
    readme: T.Optional[str] = None
    homepage: T.Optional[str] = None
    repository: T.Optional[str] = None
    license: T.Optional[str] = None
    license_file: T.Optional[str] = None
    keywords: T.List[str] = dataclasses.field(default_factory=list)
    categories: T.List[str] = dataclasses.field(default_factory=list)
    workspace: T.Optional[str] = None
    build: T.Optional[str] = None
    links: T.Optional[str] = None
    exclude: T.List[str] = dataclasses.field(default_factory=list)
    include: T.List[str] = dataclasses.field(default_factory=list)
    publish: bool = True
    metadata: T.Dict[str, T.Any] = dataclasses.field(default_factory=dict)
    default_run: T.Optional[str] = None
    autolib: bool = True
    autobins: bool = True
    autoexamples: bool = True
    autotests: bool = True
    autobenches: bool = True

    @lazy_property
    def api(self) -> str:
        return version.api(self.version)

    @classmethod
    def from_raw(cls, raw_pkg: raw.Package, workspace: T.Optional[Workspace] = None) -> Self:
        raw_ws_pkg = workspace.package if workspace else None
        return _raw_to_dataclass(raw_pkg, cls, f'Package entry {raw_pkg["name"]}', raw_ws_pkg)


@dataclasses.dataclass
class SystemDependency:

    """ Representation of a Cargo system-deps entry
        https://docs.rs/system-deps/latest/system_deps
    """

    name: str
    version: str = ''
    optional: bool = False
    feature: T.Optional[str] = None
    # TODO: convert values to dataclass
    feature_overrides: T.Dict[str, T.Dict[str, str]] = dataclasses.field(default_factory=dict)

    @lazy_property
    def meson_version(self) -> T.List[str]:
        vers = self.version.split(',') if self.version else []
        result: T.List[str] = []
        for v in vers:
            v = v.strip()
            if v[0] not in '><=':
                v = f'>={v}'
            result.append(v)
        return result

    def enabled(self, features: T.Set[str]) -> bool:
        return self.feature is None or self.feature in features

    @classmethod
    def from_raw(cls, name: str, raw: T.Union[T.Dict[str, T.Any], str]) -> Self:
        if isinstance(raw, str):
            raw = {'version': raw}
        name = raw.get('name', name)
        version = raw.get('version', '')
        optional = raw.get('optional', False)
        feature = raw.get('feature')
        # Everything else are overrides when certain features are enabled.
        feature_overrides = {k: v for k, v in raw.items() if k not in {'name', 'version', 'optional', 'feature'}}
        return cls(name, version, optional, feature, feature_overrides)


@dataclasses.dataclass
class Dependency:

    """Representation of a Cargo Dependency Entry."""

    package: str
    version: str = ''
    registry: T.Optional[str] = None
    git: T.Optional[str] = None
    branch: T.Optional[str] = None
    rev: T.Optional[str] = None
    path: T.Optional[str] = None
    optional: bool = False
    default_features: bool = True
    features: T.List[str] = dataclasses.field(default_factory=list)

    @lazy_property
    def meson_version(self) -> T.List[str]:
        return version.convert(self.version)

    @lazy_property
    def api(self) -> str:
        # Extract wanted API version from version constraints.
        api = set()
        for v in self.meson_version:
            if v.startswith(('>=', '==')):
                api.add(version.api(v[2:].strip()))
            elif v.startswith('='):
                api.add(version.api(v[1:].strip()))
        if not api:
            return ''
        elif len(api) == 1:
            return api.pop()
        else:
            raise MesonException(f'Cannot determine minimum API version from {self.version}.')

    def update_version(self, v: str) -> None:
        self.version = v
        try:
            delattr(self, 'api')
        except AttributeError:
            pass
        try:
            delattr(self, 'meson_version')
        except AttributeError:
            pass

    @T.overload
    @staticmethod
    def _depv_to_dep(depv: raw.FromWorkspace) -> raw.FromWorkspace: ...

    @T.overload
    @staticmethod
    def _depv_to_dep(depv: raw.DependencyV) -> raw.Dependency: ...

    @staticmethod
    def _depv_to_dep(depv: T.Union[raw.FromWorkspace, raw.DependencyV]) -> T.Union[raw.FromWorkspace, raw.Dependency]:
        return {'version': depv} if isinstance(depv, str) else depv

    @classmethod
    def from_raw(cls, name: str, raw_depv: T.Union[raw.FromWorkspace, raw.DependencyV], member_path: str = '', workspace: T.Optional[Workspace] = None) -> Self:
        """Create a dependency from a raw cargo dictionary or string"""
        raw_ws_dep = workspace.dependencies.get(name) if workspace else None
        raw_ws_dep = cls._depv_to_dep(raw_ws_dep or {})
        raw_dep = cls._depv_to_dep(raw_depv)

        def path_convertor(path: T.Optional[str], ws_path: T.Optional[str]) -> T.Optional[str]:
            if path:
                return path
            if ws_path:
                return os.path.relpath(ws_path, member_path)
            return None

        return _raw_to_dataclass(raw_dep, cls, f'Dependency entry {name}', raw_ws_dep,
                                 package=DefaultValue(name),
                                 path=MergeValue(path_convertor),
                                 features=MergeValue(lambda features, ws_features: (features or []) + (ws_features or [])))


@dataclasses.dataclass
class BuildTarget:

    # https://doc.rust-lang.org/cargo/reference/cargo-targets.html
    # Some default values are overridden in subclasses
    name: str
    path: str
    edition: EDITION
    test: bool = True
    doctest: bool = True
    bench: bool = True
    doc: bool = True
    harness: bool = True
    crate_type: T.List[CRATE_TYPE] = dataclasses.field(default_factory=lambda: ['bin'])
    required_features: T.List[str] = dataclasses.field(default_factory=list)
    plugin: bool = False


@dataclasses.dataclass
class Library(BuildTarget):

    """Representation of a Cargo Library Entry."""

    @classmethod
    def from_raw(cls, raw: raw.LibTarget, pkg: Package) -> Self:
        name = raw.get('name', fixup_meson_varname(pkg.name))
        # If proc_macro is True, it takes precedence and sets crate_type to proc-macro
        proc_macro = raw.get('proc-macro', False)
        return _raw_to_dataclass(raw, cls, f'Library entry {name}',
                                 ignored_fields=['proc-macro'],
                                 name=DefaultValue(name),
                                 path=DefaultValue('src/lib.rs'),
                                 edition=DefaultValue(pkg.edition),
                                 crate_type=ConvertValue(lambda x: ['proc-macro'] if proc_macro else x,
                                                         ['proc-macro'] if proc_macro else ['lib']))


@dataclasses.dataclass
class Binary(BuildTarget):

    """Representation of a Cargo Bin Entry."""

    @classmethod
    def from_raw(cls, raw: raw.BuildTarget, pkg: Package) -> Self:
        name = raw["name"]
        return _raw_to_dataclass(raw, cls, f'Binary entry {name}',
                                 path=DefaultValue('src/main.rs'),
                                 edition=DefaultValue(pkg.edition))


@dataclasses.dataclass
class Test(BuildTarget):

    """Representation of a Cargo Test Entry."""

    @classmethod
    def from_raw(cls, raw: raw.BuildTarget, pkg: Package) -> Self:
        name = raw["name"]
        return _raw_to_dataclass(raw, cls, f'Test entry {name}',
                                 path=DefaultValue(f'tests/{name}.rs'),
                                 edition=DefaultValue(pkg.edition),
                                 bench=DefaultValue(False),
                                 doc=DefaultValue(False))


@dataclasses.dataclass
class Benchmark(BuildTarget):

    """Representation of a Cargo Benchmark Entry."""

    @classmethod
    def from_raw(cls, raw: raw.BuildTarget, pkg: Package) -> Self:
        name = raw["name"]
        return _raw_to_dataclass(raw, cls, f'Benchmark entry {name}',
                                 path=DefaultValue(f'benches/{name}.rs'),
                                 edition=DefaultValue(pkg.edition),
                                 test=DefaultValue(False),
                                 doc=DefaultValue(False))


@dataclasses.dataclass
class Example(BuildTarget):

    """Representation of a Cargo Example Entry."""

    @classmethod
    def from_raw(cls, raw: raw.BuildTarget, pkg: Package) -> Self:
        name = raw["name"]
        return _raw_to_dataclass(raw, cls, f'Example entry {name}',
                                 path=DefaultValue(f'examples/{name}.rs'),
                                 edition=DefaultValue(pkg.edition),
                                 test=DefaultValue(False),
                                 bench=DefaultValue(False),
                                 doc=DefaultValue(False))


@dataclasses.dataclass
class Lint:

    """Cargo Lint definition.
    """

    name: str
    level: LINT_LEVEL
    priority: int
    check_cfg: T.Optional[T.List[str]]

    @classmethod
    def from_raw(cls, r: T.Union[raw.FromWorkspace, T.Dict[str, T.Dict[str, raw.LintV]]]) -> T.List[Lint]:
        r = T.cast('T.Dict[str, T.Dict[str, raw.LintV]]', r)
        lints: T.Dict[str, Lint] = {}
        for tool, raw_lints in r.items():
            prefix = '' if tool == 'rust' else f'{tool}::'
            for name, settings in raw_lints.items():
                name = prefix + name
                if isinstance(settings, str):
                    settings = T.cast('raw.Lint', {'level': settings})
                check_cfg = None
                if name == 'unexpected_cfgs':
                    check_cfg = settings.get('check-cfg', [])
                lints[name] = Lint(name=name,
                                   level=settings['level'],
                                   priority=settings.get('priority', 0),
                                   check_cfg=check_cfg)

        lints_final = list(lints.values())
        lints_final.sort(key=lambda x: x.priority)
        return lints_final

    def to_arguments(self, check_cfg: bool) -> T.List[str]:
        if self.level == "deny":
            flag = "-D"
        elif self.level == "allow":
            flag = "-A"
        elif self.level == "warn":
            flag = "-W"
        elif self.level == "forbid":
            flag = "-F"
        else:
            raise MesonException(f"invalid level {self.level!r} for {self.name}")
        args = [flag, self.name]
        if check_cfg and self.check_cfg:
            for arg in self.check_cfg:
                args.append('--check-cfg')
                args.append(arg)
        return args


@dataclasses.dataclass
class Manifest:

    """Cargo Manifest definition.

    Most of these values map up to the Cargo Manifest, but with default values
    if not provided.

    Cargo subprojects can contain what Meson wants to treat as multiple,
    interdependent, subprojects.

    :param path: the path within the cargo subproject.
    """

    package: Package
    dependencies: T.Dict[str, Dependency] = dataclasses.field(default_factory=dict)
    dev_dependencies: T.Dict[str, Dependency] = dataclasses.field(default_factory=dict)
    build_dependencies: T.Dict[str, Dependency] = dataclasses.field(default_factory=dict)
    lib: T.Optional[Library] = None
    bin: T.Dict[str, Binary] = dataclasses.field(default_factory=dict)
    test: T.List[Test] = dataclasses.field(default_factory=list)
    bench: T.List[Benchmark] = dataclasses.field(default_factory=list)
    example: T.List[Example] = dataclasses.field(default_factory=list)
    features: T.Dict[str, T.List[str]] = dataclasses.field(default_factory=dict)
    target: T.Dict[str, T.Dict[str, Dependency]] = dataclasses.field(default_factory=dict)
    lints: T.List[Lint] = dataclasses.field(default_factory=list)

    # missing: profile

    def __post_init__(self) -> None:
        self.features.setdefault('default', [])

    @lazy_property
    def system_dependencies(self) -> T.Dict[str, SystemDependency]:
        return {k: SystemDependency.from_raw(k, v) for k, v in self.package.metadata.get('system-deps', {}).items()}

    @classmethod
    def from_raw(cls, raw: raw.Manifest, path: str, workspace: T.Optional[Workspace] = None, member_path: str = '') -> Self:
        pkg = Package.from_raw(raw['package'], workspace)

        autolib = None
        if pkg.autolib and os.path.exists(os.path.join(path, 'src/lib.rs')):
            autolib = Library.from_raw({}, pkg)

        def _discover_targets(subdir: str) -> T.Generator[T.Tuple[str, str], None, None]:
            """Discover .rs files in a subdirectory and yield (name, path) tuples."""
            target_dir = os.path.join(path, subdir)
            if os.path.isdir(target_dir):
                for entry in os.listdir(target_dir):
                    if entry.endswith('.rs'):
                        target_name = entry[:-3]  # Remove .rs extension
                        yield target_name, f'{subdir}/{entry}'

        autobins: T.Dict[str, Binary] = {}
        if pkg.autobins:
            # Check for default binary (src/main.rs)
            if os.path.exists(os.path.join(path, 'src/main.rs')):
                autobins[pkg.name] = Binary.from_raw({'name': pkg.name, 'path': 'src/main.rs'}, pkg)
            # Add additional binaries from src/bin/
            for bin_name, bin_path in _discover_targets('src/bin'):
                if bin_name in autobins:
                    raise MesonException(f'Binary target {bin_name!r} is defined more than once '
                                         f'({autobins[bin_name].path} and {bin_path})')
                autobins[bin_name] = Binary.from_raw({'name': bin_name, 'path': bin_path}, pkg)

        def dependencies_from_raw(x: T.Dict[str, T.Any]) -> T.Dict[str, Dependency]:
            return {k: Dependency.from_raw(k, v, member_path, workspace) for k, v in x.items()}

        return _raw_to_dataclass(raw, cls, f'Cargo.toml package {pkg.name}',
                                 raw_from_workspace=workspace.inheritable if workspace else None,
                                 ignored_fields=['badges', 'workspace'],
                                 package=ConvertValue(lambda _: pkg),
                                 dependencies=ConvertValue(dependencies_from_raw),
                                 dev_dependencies=ConvertValue(dependencies_from_raw),
                                 build_dependencies=ConvertValue(dependencies_from_raw),
                                 lints=ConvertValue(Lint.from_raw),
                                 lib=ConvertValue(lambda x: Library.from_raw(x, pkg), default=autolib),
                                 bin=DictMergeValue(lambda x: [Binary.from_raw(b, pkg) for b in x],
                                                    merge_key=lambda x: x.path,
                                                    out_key=lambda x: x.name,
                                                    base=autobins),
                                 test=ConvertValue(lambda x: [Test.from_raw(b, pkg) for b in x]),
                                 bench=ConvertValue(lambda x: [Benchmark.from_raw(b, pkg) for b in x]),
                                 example=ConvertValue(lambda x: [Example.from_raw(b, pkg) for b in x]),
                                 target=ConvertValue(lambda x: {k: dependencies_from_raw(v.get('dependencies', {})) for k, v in x.items()}))


@dataclasses.dataclass
class Workspace:

    """Cargo Workspace definition.
    """

    resolver: str = dataclasses.field(default_factory=lambda: '2')
    members: T.List[str] = dataclasses.field(default_factory=list)
    exclude: T.List[str] = dataclasses.field(default_factory=list)
    default_members: T.List[str] = dataclasses.field(default_factory=list)

    # inheritable settings are kept in raw format, for use with _raw_to_dataclass
    package: T.Optional[raw.Package] = None
    dependencies: T.Dict[str, raw.Dependency] = dataclasses.field(default_factory=dict)
    lints: T.Dict[str, T.Dict[str, raw.LintV]] = dataclasses.field(default_factory=dict)
    metadata: T.Dict[str, T.Any] = dataclasses.field(default_factory=dict)

    # A workspace can also have a root package.
    root_package: T.Optional[Manifest] = None

    @lazy_property
    def inheritable(self) -> T.Dict[str, object]:
        # the whole lints table is inherited.  Do not add package, dependencies
        # etc. because they can only be inherited a field at a time.
        return {
            'lints': self.lints,
        }

    @classmethod
    def from_raw(cls, raw: raw.Manifest, path: str) -> Self:
        ws = _raw_to_dataclass(raw['workspace'], cls, 'Workspace')
        if 'package' in raw:
            ws.root_package = Manifest.from_raw(raw, path, ws, '.')
        if not ws.default_members:
            ws.default_members = ['.'] if ws.root_package else ws.members
        return ws


@dataclasses.dataclass
class CargoLockPackage:

    """A description of a package in the Cargo.lock file format."""

    name: str
    version: str
    source: T.Optional[str] = None
    checksum: T.Optional[str] = None
    dependencies: T.List[str] = dataclasses.field(default_factory=list)

    @lazy_property
    def api(self) -> str:
        return version.api(self.version)

    @lazy_property
    def subproject(self) -> str:
        return f'{self.name}-{self.api}-rs'

    @classmethod
    def from_raw(cls, raw: raw.CargoLockPackage) -> Self:
        return _raw_to_dataclass(raw, cls, 'Cargo.lock package')


@dataclasses.dataclass
class CargoLock:

    """A description of the Cargo.lock file format."""

    version: int = 1
    package: T.List[CargoLockPackage] = dataclasses.field(default_factory=list)
    metadata: T.Dict[str, str] = dataclasses.field(default_factory=dict)
    wraps: T.Dict[str, PackageDefinition] = dataclasses.field(default_factory=dict)

    def named(self, name: str) -> T.Sequence[CargoLockPackage]:
        return self._versions[name]

    @lazy_property
    def _versions(self) -> T.Dict[str, T.List[CargoLockPackage]]:
        versions = collections.defaultdict(list)
        for pkg in self.package:
            versions[pkg.name].append(pkg)
        for pkg_versions in versions.values():
            pkg_versions.sort(reverse=True, key=lambda pkg: Version(pkg.version))
        return versions

    @classmethod
    def from_raw(cls, raw: raw.CargoLock) -> Self:
        return _raw_to_dataclass(raw, cls, 'Cargo.lock',
                                 package=ConvertValue(lambda x: [CargoLockPackage.from_raw(p) for p in x]))



# SPDX-License-Identifier: Apache-2.0
# Copyright 2021 The Meson development team

from .generatorbase import GeneratorBase
from .model import ReferenceManual, Object, Function, DataTypeInfo, Type, ObjectType

from mesonbuild import mlog
import typing as T

def my_nested() -> T.ContextManager[None]:
    prefix = '|' * mlog.get_log_depth()
    return mlog.nested(prefix)

class GeneratorPrint(GeneratorBase):
    def _types_to_string(self, typ: Type) -> str:
        def _data_type_to_str(dt: DataTypeInfo) -> str:
            if dt.holds:
                return f'{dt.data_type.name}[{self._types_to_string(dt.holds)}]'
            return dt.data_type.name
        return ' | '.join([_data_type_to_str(x) for x in typ.resolved])

    def _generate_function(self, func: Function) -> None:
        mlog.log()
        mlog.log('Function', mlog.bold(func.name))
        with my_nested():
            desc = func.description
            if '\n' in desc:
                desc = desc[:desc.index('\n')]
            mlog.log('Description:', mlog.bold(desc))
            mlog.log('Return type:', mlog.bold(self._types_to_string(func.returns)))
            mlog.log('Pos args:   ', mlog.bold(str([x.name for x in func.posargs])))
            mlog.log('Opt args:   ', mlog.bold(str([x.name for x in func.optargs])))
            mlog.log('Varargs:    ', mlog.bold(func.varargs.name if func.varargs is not None else 'null'))
            mlog.log('Kwargs:     ', mlog.bold(str(list(func.kwargs.keys()))))

    def _generate_object(self, obj: Object) -> None:
        tags = []
        tags += [{
            ObjectType.ELEMENTARY: mlog.yellow('[elementary]'),
            ObjectType.BUILTIN: mlog.green('[builtin]'),
            ObjectType.MODULE: mlog.blue('[module]'),
            ObjectType.RETURNED: mlog.cyan('[returned]'),
        }[obj.obj_type]]
        if obj.is_container:
            tags += [mlog.red('[container]')]
        mlog.log()
        mlog.log('Object', mlog.bold(obj.name), *tags)
        with my_nested():
            desc = obj.description
            if '\n' in desc:
                desc = desc[:desc.index('\n')]
            mlog.log('Description:', mlog.bold(desc))
            mlog.log('Returned by:', mlog.bold(str([x.name for x in obj.returned_by])))
            mlog.log('Methods:')
            with my_nested():
                for m in obj.methods:
                    self._generate_function(m)

    def generate(self) -> None:
        mlog.log('\n\n', mlog.bold('=== Functions ==='), '\n')
        for f in self.functions:
            self._generate_function(f)
        mlog.log('\n\n', mlog.bold('=== Elementary ==='), '\n')
        for obj in self.elementary:
            self._generate_object(obj)
        mlog.log('\n\n', mlog.bold('=== Builtins ==='), '\n')
        for obj in self.builtins:
            self._generate_object(obj)
        mlog.log('\n\n', mlog.bold('=== Returned objects ==='), '\n')
        for obj in self.returned:
            self._generate_object(obj)
        mlog.log('\n\n', mlog.bold('=== Modules ==='), '\n')
        for obj in self.modules:
            self._generate_object(obj)
            for mod_obj in self.extract_returned_by_module(obj):
                self._generate_object(mod_obj)



#!/usr/bin/env python3

import shutil, sys, subprocess, argparse, pathlib
import platform

parser = argparse.ArgumentParser()

parser.add_argument('--private-dir', required=True)
parser.add_argument('-o', required=True)
parser.add_argument('cmparr', nargs='+')

contents = '''#include<stdio.h>

void flob(void) {
    printf("Now flobbing.\\n");
}
'''

def get_pic_args():
    platname = platform.system().lower()
    if platname in ['windows', 'darwin'] or sys.platform == 'cygwin':
        return []
    return ['-fPIC']

def generate_lib_gnulike(outfile, c_file, private_dir, compiler_array):
    if shutil.which('ar'):
        static_linker = 'ar'
    elif shutil.which('llvm-ar'):
        static_linker = 'llvm-ar'
    elif shutil.which('gcc-ar'):
        static_linker = 'gcc-ar'
    else:
        sys.exit('Could not detect a static linker.')
    o_file = c_file.with_suffix('.o')
    compile_cmd = compiler_array + ['-c', '-g', '-O2', '-o', str(o_file), str(c_file)]
    compile_cmd += get_pic_args()
    subprocess.check_call(compile_cmd)
    out_file = pathlib.Path(outfile)
    if out_file.exists():
        out_file.unlink()
    link_cmd = [static_linker, 'csr', outfile, str(o_file)]
    subprocess.check_call(link_cmd)
    return 0


def generate_lib_msvc(outfile, c_file, private_dir, compiler_array):
    static_linker = 'lib'
    o_file = c_file.with_suffix('.obj')
    compile_cmd = compiler_array + ['/MDd',
                                    '/nologo',
                                    '/ZI',
                                    '/Ob0',
                                    '/Od',
                                    '/c',
                                    '/Fo' + str(o_file),
                                    str(c_file)]
    subprocess.check_call(compile_cmd)
    out_file = pathlib.Path(outfile)
    if out_file.exists():
        out_file.unlink()
    link_cmd = [static_linker,
                '/nologo',
                '/OUT:' + str(outfile),
                str(o_file)]
    subprocess.check_call(link_cmd)
    return 0

def generate_lib(outfile, private_dir, compiler_array):
    private_dir = pathlib.Path(private_dir)
    if not private_dir.exists():
        private_dir.mkdir()
    c_file = private_dir / 'flob.c'
    c_file.write_text(contents)
    for i in compiler_array:
        if (i.endswith('cl') or i.endswith('cl.exe')) and 'clang-cl' not in i:
            return generate_lib_msvc(outfile, c_file, private_dir, compiler_array)
    return generate_lib_gnulike(outfile, c_file, private_dir, compiler_array)

if __name__ == '__main__':
    options = parser.parse_args()
    sys.exit(generate_lib(options.o, options.private_dir, options.cmparr))



# SPDX-License-Identifier: Apache-2.0
# Copyright 2021 The Meson development team
from __future__ import annotations

from pathlib import Path
import json
import re

from .generatorbase import GeneratorBase
from . import jsonschema as J
from .model import (
    ReferenceManual,
    Function,
    Object,
    Type,

    PosArg,
    VarArgs,
    Kwarg,
)

import typing as T

class GeneratorJSON(GeneratorBase):
    def __init__(self, manual: ReferenceManual, out: Path, enable_modules: bool) -> None:
        super().__init__(manual)
        self.out = out
        self.enable_modules = enable_modules

    def _generate_type(self, typ: Type) -> T.List[J.Type]:
        return [
            {
                'obj': x.data_type.name,
                'holds': self._generate_type(x.holds) if x.holds else [],
            }
            for x in typ.resolved
        ]

    def _generate_type_str(self, typ: Type) -> str:
        # Remove all whitespaces
        return re.sub(r'[ \n\r\t]', '', typ.raw)

    def _generate_arg(self, arg: T.Union[PosArg, VarArgs, Kwarg], isOptarg: bool = False) -> J.Argument:
        return {
            'name': arg.name,
            'description': arg.description,
            'since': arg.since if arg.since else None,
            'deprecated': arg.deprecated if arg.deprecated else None,
            'type': self._generate_type(arg.type),
            'type_str': self._generate_type_str(arg.type),
            'required': arg.required if isinstance(arg, Kwarg) else not isOptarg and not isinstance(arg, VarArgs),
            'default': arg.default if isinstance(arg, (PosArg, Kwarg)) else None,
            'min_varargs': arg.min_varargs if isinstance(arg, VarArgs) and arg.min_varargs > 0 else None,
            'max_varargs': arg.max_varargs if isinstance(arg, VarArgs) and arg.max_varargs > 0 else None,

            # Not yet supported
            'notes': [],
            'warnings': [],
        }

    def _generate_function(self, func: Function) -> J.Function:
        return {
            'name': func.name,
            'description': func.description,
            'since': func.since if func.since else None,
            'deprecated': func.deprecated if func.deprecated else None,
            'notes': func.notes,
            'warnings': func.warnings,
            'example': func.example if func.example else None,
            'returns': self._generate_type(func.returns),
            'returns_str': self._generate_type_str(func.returns),
            'posargs': {x.name: self._generate_arg(x) for x in func.posargs},
            'optargs': {x.name: self._generate_arg(x, True) for x in func.optargs},
            'kwargs': {x.name: self._generate_arg(x) for x in self.sorted_and_filtered(list(func.kwargs.values()))},
            'varargs': self._generate_arg(func.varargs) if func.varargs else None,
            'arg_flattening': func.arg_flattening,
        }

    def _generate_objects(self, obj: Object) -> J.Object:
        return {
            'name': obj.name,
            'description': obj.description,
            'since': obj.since if obj.since else None,
            'deprecated': obj.deprecated if obj.deprecated else None,
            'notes': obj.notes,
            'warnings': obj.warnings,
            'defined_by_module': obj.defined_by_module.name if obj.defined_by_module else None,
            'object_type': obj.obj_type.name,
            'is_container': obj.is_container,
            'example': obj.example if obj.example else None,
            'extends': obj.extends if obj.extends else None,
            'returned_by': [x.name for x in self.sorted_and_filtered(obj.returned_by)],
            'extended_by': [x.name for x in self.sorted_and_filtered(obj.extended_by)],
            'methods': {x.name: self._generate_function(x) for x in self.sorted_and_filtered(obj.methods)},
        }

    def generate(self) -> None:
        data: J.Root = {
            'version_major': J.VERSION_MAJOR,
            'version_minor': J.VERSION_MINOR,
            'meson_version': self._extract_meson_version(),
            'functions': {x.name: self._generate_function(x) for x in self.sorted_and_filtered(self.functions)},
            'objects': {x.name: self._generate_objects(x) for x in self.sorted_and_filtered(self.objects)},
            'objects_by_type': {
                'elementary': [x.name for x in self.elementary],
                'builtins': [x.name for x in self.builtins],
                'returned': [x.name for x in self.returned],
                'modules': {
                    x.name: [y.name for y in self.sorted_and_filtered(self.extract_returned_by_module(x))]
                    for x in self.modules
                },
            },
        }

        self.out.write_text(json.dumps(data), encoding='utf-8')



#!/usr/bin/env python3

import argparse
import os


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument('builddir')
    args = parser.parse_args()

    for _, _, files in os.walk(args.builddir):
        if 'main-unique.rs' in files:
            exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import os

assert 'ENV_A' in os.environ
assert 'ENV_B' in os.environ
assert 'ENV_C' in os.environ

print('ENV_A is', os.environ['ENV_A'])
print('ENV_B is', os.environ['ENV_B'])
print('ENV_C is', os.environ['ENV_C'])



# SPDX-License-Identifier: Apache-2.0
# Copyright © 2020-2025 Intel Corporation

from __future__ import annotations
import itertools
import os
import re
import textwrap
import typing as T

from mesonbuild.interpreterbase.decorators import FeatureNew

from . import ExtensionModule, ModuleReturnValue, ModuleInfo, ModuleObject
from .. import mesonlib, mlog
from ..build import (BothLibraries, BuildTarget, CustomTargetIndex, Executable, ExtractedObjects, GeneratedList,
                     CustomTarget, InvalidArguments, Jar, StructuredSources, SharedLibrary, StaticLibrary,
                     SharedModule)
from ..compilers.compilers import are_asserts_disabled_for_subproject, lang_suffixes
from ..compilers.rust import parse_target, RustSystemDependency
from ..dependencies import Dependency
from ..interpreter.type_checking import (
    DEPENDENCIES_KW, LINK_WITH_KW, LINK_WHOLE_KW, SHARED_LIB_KWS, TEST_KWS, TEST_KWS_NO_ARGS,
    OUTPUT_KW, INCLUDE_DIRECTORIES, SOURCES_VARARGS, NATIVE_KW, NoneType, in_set_validator,
    EXECUTABLE_KWS, LIBRARY_KWS, SHARED_MOD_KWS, _BASE_LANG_KW
)
from ..interpreterbase import ContainerTypeInfo, InterpreterException, KwargInfo, typed_kwargs, typed_pos_args, noKwargs, noPosargs, permittedKwargs
from ..interpreter.interpreterobjects import Doctest
from ..mesonlib import (is_parent_path, File, MachineChoice, MesonException, PerMachine)
from ..programs import ExternalProgram, NonExistingExternalProgram

if T.TYPE_CHECKING:
    from . import ModuleState
    from .. import cargo
    from ..build import BuildTargetTypes, ExecutableKeywordArguments, IncludeDirs, LibTypes
    from ..cargo.interpreter import RUST_ABI
    from ..compilers.compilers import Language
    from ..compilers.rust import RustCompiler
    from ..dependencies import ExternalLibrary
    from ..interpreter import Interpreter
    from ..interpreter import kwargs as _kwargs
    from ..interpreter.interpreter import SourceInputs, SourceOutputs
    from ..interpreter.interpreterobjects import Test
    from ..interpreterbase import TYPE_kwargs
    from ..programs import Program
    from ..interpreter.type_checking import SourcesVarargsType

    from typing_extensions import TypedDict, Literal

    ArgsType = T.TypeVar('ArgsType')

    class FuncRustTest(_kwargs.BaseTest, T.Generic[ArgsType]):
        args: T.List[ArgsType]
        dependencies: T.List[T.Union[Dependency, ExternalLibrary]]
        is_parallel: bool
        link_with: T.List[LibTypes]
        link_whole: T.List[T.Union[StaticLibrary, CustomTarget, CustomTargetIndex]]
        rust_args: T.List[str]

    FuncTest = FuncRustTest[_kwargs.TestArgs]
    FuncDoctest = FuncRustTest[str]

    class FuncBindgen(TypedDict):

        args: T.List[str]
        c_args: T.List[str]
        include_directories: T.List[IncludeDirs]
        input: T.List[SourceInputs]
        output: str
        output_inline_wrapper: str
        dependencies: T.List[T.Union[Dependency, ExternalLibrary]]
        language: T.Optional[Literal['c', 'cpp']]
        bindgen_version: T.List[str]

    class FuncWorkspace(TypedDict):
        default_features: T.Optional[bool]
        features: T.List[str]

    class FuncDependency(TypedDict):
        rust_abi: T.Optional[RUST_ABI]

    class RustPackageDependencies(TypedDict):
        dependencies: bool
        dev_dependencies: bool
        system_dependencies: bool

    class RustPackageExecutable(_kwargs.Executable):
        pass

    class RustPackageLibrary(_kwargs.Library):
        pass

RUST_TEST_KWS: T.List[KwargInfo] = [
     KwargInfo(
         'rust_args',
         ContainerTypeInfo(list, str),
         listify=True,
         default=[],
         since='1.2.0',
     ),
     KwargInfo('is_parallel', bool, default=False),
]

# The native argument should be passed to the "package" method
EXECUTABLE_KWS_NO_NATIVE = [kwi for kwi in EXECUTABLE_KWS if kwi.name != 'native']
LIBRARY_KWS_NO_NATIVE = [kwi for kwi in LIBRARY_KWS if kwi.name != 'native']
SHARED_LIB_KWS_NO_NATIVE = [kwi for kwi in SHARED_LIB_KWS if kwi.name != 'native']
SHARED_MOD_KWS_NO_NATIVE = [kwi for kwi in SHARED_MOD_KWS if kwi.name != 'native']


def no_spaces_validator(arg: T.Optional[T.Union[str, T.List]]) -> T.Optional[str]:
    if any(bool(re.search(r'\s', x)) for x in arg):
        return 'must not contain spaces due to limitations of rustdoc'
    return None

def dep_to_system_dependency(dep: Dependency, depname: str) -> Dependency:
    if not dep.found():
        return dep
    if not depname:
        if not dep.name:
            raise MesonException("rust.to_system_dependency() called with an unnamed dependency and no explicit name")
        depname = dep.name
    depname = re.sub(r'[^a-zA-Z0-9]', '_', depname)
    rust_args = ['--cfg', f'system_deps_have_{depname}']
    return RustSystemDependency(dep.version, compile_args=rust_args, ext_deps=[dep], name=dep.name)

class RustWorkspace(ModuleObject):
    """Represents a Rust workspace, controlling the build of packages
       recorded in a Cargo.lock file."""

    def __init__(self, interpreter: Interpreter, ws: cargo.WorkspaceState) -> None:
        super().__init__()
        self.interpreter = interpreter
        self.ws = ws
        self.methods.update({
            'packages': self.packages_method,
            'package': self.package_method,
            'subproject': self.subproject_method,
        })

    @property
    def subdir(self) -> str:
        return self.ws.subdir

    @noPosargs
    @noKwargs
    def packages_method(self, state: ModuleState, args: T.List, kwargs: TYPE_kwargs) -> T.List[str]:
        """Returns list of package names in workspace."""
        package_names = [pkg.manifest.package.name for pkg in self.ws.packages.values()]
        return sorted(package_names)

    @typed_pos_args('workspace.package', optargs=[str])
    def package_method(self, state: 'ModuleState', args: T.List, kwargs: TYPE_kwargs) -> RustPackage:
        """Returns a package object."""
        package_name = args[0] if args else None
        return RustPackage(self, self.interpreter.cargo.load_package(self.ws, package_name))

    def _do_subproject(self, pkg: cargo.PackageState) -> None:
        kw: _kwargs.DoSubproject = {
            'required': True,
            'version': None,
            'options': None,
            'cmake_options': [],
            'default_options': {},
        }
        subp_name = pkg.get_subproject_name()
        self.interpreter.do_subproject(subp_name, kw, force_method='cargo')

    @typed_pos_args('workspace.subproject', str, optargs=[str])
    @noKwargs
    def subproject_method(self, state: ModuleState, args: T.Tuple[str, T.Optional[str]], kwargs: TYPE_kwargs) -> RustSubproject:
        """Returns a package object for a subproject package."""
        package_name = args[0]
        pkg = self.interpreter.cargo.resolve_package(package_name, args[1] or '')
        if pkg is None:
            if args[1]:
                raise MesonException(f'No version of cargo package "{package_name}" provides API {args[1]}')
            else:
                raise MesonException(f'Cargo package "{package_name}" not available')

        self._do_subproject(pkg)
        return RustSubproject(self, pkg)


class RustCrate(ModuleObject):
    """Abstract base class for Rust crate representations."""

    def __init__(self, rust_ws: RustWorkspace, package: cargo.PackageState) -> None:
        super().__init__()
        self.rust_ws = rust_ws
        self.package = package
        self.methods.update({
            'all_features': self.all_features_method,
            'api': self.api_method,
            'features': self.features_method,
            'name': self.name_method,
            'version': self.version_method,
            'rust_args': self.rust_args_method,
            'env': self.env_method, # type: ignore[dict-item]
            'rust_dependency_map': self.rust_dependency_map_method, # type: ignore[dict-item]
        })

    @noPosargs
    @noKwargs
    def name_method(self, state: ModuleState, args: T.List, kwargs: TYPE_kwargs) -> str:
        """Returns the name of the package."""
        return self.package.manifest.package.name

    @noPosargs
    @noKwargs
    def api_method(self, state: ModuleState, args: T.List, kwargs: TYPE_kwargs) -> str:
        """Returns the API version of the package."""
        return self.package.manifest.package.api

    @noPosargs
    @noKwargs
    def version_method(self, state: ModuleState, args: T.List, kwargs: TYPE_kwargs) -> str:
        """Returns the version of the package."""
        return self.package.manifest.package.version

    @noPosargs
    @noKwargs
    def all_features_method(self, state: ModuleState, args: T.List, kwargs: TYPE_kwargs) -> T.List[str]:
        """Returns all features for specific package."""
        return sorted(list(self.package.manifest.features.keys()))

    @noPosargs
    @noKwargs
    def features_method(self, state: ModuleState, args: T.List, kwargs: TYPE_kwargs) -> T.List[str]:
        """Returns chosen features for specific package."""
        return sorted(list(self.package.cfg.features))

    @noPosargs
    @noKwargs
    def rust_args_method(self, state: ModuleState, args: T.List, kwargs: TYPE_kwargs) -> T.List[str]:
        """Returns rustc arguments for this package."""
        return self.package.get_rustc_args(state.environment, state.subdir, mesonlib.MachineChoice.HOST)

    @noPosargs
    @noKwargs
    def env_method(self, state: ModuleState, args: T.List, kwargs: TYPE_kwargs) -> T.Dict[str, str]:
        """Returns environment variables for this package."""
        return self.package.get_env_dict(state.environment, state.subdir)

    @noPosargs
    @noKwargs
    def rust_dependency_map_method(self, state: ModuleState, args: T.List, kwargs: TYPE_kwargs) -> T.Dict[str, str]:
        """Returns rust dependency mapping for this package."""
        return self.package.cfg.get_dependency_map(self.package.manifest)


class RustPackage(RustCrate):
    """Represents a Rust package within a workspace."""

    def __init__(self, rust_ws: RustWorkspace, package: cargo.PackageState) -> None:
        super().__init__(rust_ws, package)
        self.methods.update({
            'dependencies': self.dependencies_method,
            'library': self.library_method,
            'proc_macro': self.proc_macro_method,
            'shared_module': self.shared_module_method,
            'executable': self.executable_method,
            'override_dependency': self.override_dependency_method,
        })

    def _dependencies_method(self, state: ModuleState, kwargs: RustPackageDependencies,
                             for_machine: MachineChoice) -> T.List[Dependency]:
        dependencies: T.List[Dependency] = []
        cfg = self.package.cfg

        if kwargs['dependencies']:
            for dep_key, dep_pkg in cfg.dep_packages.items():
                if dep_pkg.manifest.lib:
                    if dep_pkg.ws_subdir != self.rust_ws.subdir or \
                        is_parent_path(os.path.join(self.rust_ws.subdir, state.subproject_dir),
                                       dep_pkg.path):
                        self.rust_ws._do_subproject(dep_pkg)
                    # Get the dependency name for this package (rust or proc-macro ABI)
                    depname = dep_pkg.get_rust_dependency_name()
                    dependency = state.overridden_dependency(depname, for_machine)
                    dependencies.append(dependency)

        if kwargs['dev_dependencies']:
            raise MesonException('dev_dependencies is not implemented yet')

        if kwargs['system_dependencies']:
            for name, sys_dep in self.package.manifest.system_dependencies.items():
                if sys_dep.enabled(cfg.features):
                    # System dependencies use the original dependency name from Cargo.toml
                    dependency = state.dependency(sys_dep.name, native=(for_machine == MachineChoice.BUILD),
                                                  required=not sys_dep.optional,
                                                  wanted=sys_dep.meson_version)
                    dependencies.append(dep_to_system_dependency(dependency, name))

        return dependencies

    @noPosargs
    @typed_kwargs('package.dependencies',
                  KwargInfo('dependencies', bool, default=True),
                  KwargInfo('dev_dependencies', bool, default=False),
                  KwargInfo('system_dependencies', bool, default=True))
    def dependencies_method(self, state: ModuleState, args: T.List, kwargs: RustPackageDependencies) -> T.List[Dependency]:
        """Returns the dependencies for this package."""
        return self._dependencies_method(state, kwargs, MachineChoice.HOST)

    @staticmethod
    def validate_pos_args(name: str, args: T.Tuple[
            T.Optional[T.Union[str, StructuredSources]],
            T.Optional[StructuredSources]]) -> T.Tuple[T.Optional[str], T.Optional[StructuredSources]]:
        if isinstance(args[0], str):
            return args[0], args[1]
        if args[1] is not None:
            raise MesonException(f"{name} only accepts one StructuredSources parameter")
        return None, args[0]

    def merge_kw_args(self, state: ModuleState, kwargs: T.Union[RustPackageExecutable, RustPackageLibrary]) -> None:
        kwargs.setdefault('native', MachineChoice.HOST)

        deps = kwargs['dependencies']
        kwargs['dependencies'] = self._dependencies_method(state, {
            'dependencies': True,
            'dev_dependencies': False,
            'system_dependencies': True,
        }, kwargs['native'])
        kwargs['dependencies'].extend(deps)

        depmap = kwargs['rust_dependency_map']
        kwargs['rust_dependency_map'] = \
            self.package.cfg.get_dependency_map(self.package.manifest)
        kwargs['rust_dependency_map'].update(depmap)

        rust_args = kwargs['rust_args']
        kwargs['rust_args'] = \
            self.package.get_rustc_args(state.environment, state.subdir, kwargs['native'])
        kwargs['rust_args'].extend(rust_args)

        kwargs['override_options'].setdefault('rust_std', self.package.manifest.package.edition)

    def _library_method(self, state: ModuleState, args: T.Tuple[
            T.Optional[T.Union[str, StructuredSources]],
            T.Optional[StructuredSources]], kwargs: RustPackageLibrary,
            static: bool, shared: bool,
            shared_mod: bool = False) -> T.Union[BothLibraries, SharedLibrary, StaticLibrary]:
        tgt_args = self.validate_pos_args('package.library', args)
        if not self.package.manifest.lib:
            raise MesonException("no [lib] section in Cargo package")

        sources: T.Union[StructuredSources, str]
        tgt_name, sources = tgt_args
        if not tgt_name:
            rust_abi: RUST_ABI
            if kwargs['rust_crate_type'] is not None:
                rust_abi = 'rust' if kwargs['rust_crate_type'] in {'lib', 'rlib', 'dylib', 'proc-macro'} else 'c'
            else:
                rust_abi = kwargs['rust_abi']
            tgt_name = self.package.library_name(rust_abi)
        if not sources:
            sources = os.path.relpath(os.path.join(self.package.path, self.package.manifest.lib.path),
                                      state.subdir)

        lib_args: T.Tuple[str, SourcesVarargsType] = (tgt_name, [sources])
        self.merge_kw_args(state, kwargs)

        if shared_mod:
            return state._interpreter.build_target(state.current_node, lib_args,
                                                   T.cast('_kwargs.SharedModule', kwargs),
                                                   SharedModule)

        if static and shared:
            return state._interpreter.build_both_libraries(state.current_node, lib_args, kwargs)
        elif shared:
            return state._interpreter.build_target(state.current_node, lib_args,
                                                   T.cast('_kwargs.SharedLibrary', kwargs),
                                                   SharedLibrary)
        else:
            return state._interpreter.build_target(state.current_node, lib_args,
                                                   T.cast('_kwargs.StaticLibrary', kwargs),
                                                   StaticLibrary)

    def _proc_macro_method(self, state: 'ModuleState', args: T.Tuple[
            T.Optional[T.Union[str, StructuredSources]],
            T.Optional[StructuredSources]], kwargs: RustPackageLibrary) -> SharedLibrary:
        kwargs['native'] = MachineChoice.BUILD
        kwargs['rust_abi'] = None
        kwargs['rust_crate_type'] = 'proc-macro'
        kwargs['rust_args'] = kwargs['rust_args'] + ['--extern', 'proc_macro']
        result = self._library_method(state, args, kwargs, shared=True, static=False)
        return T.cast('SharedLibrary', result)

    @typed_pos_args('package.override_dependency', Dependency)
    @typed_kwargs('package.override_dependency',
                  KwargInfo('rust_abi', (str, NoneType), default=None, validator=in_set_validator({'rust', 'c', 'proc-macro'})))
    def override_dependency_method(self, state: ModuleState, args: T.Tuple[Dependency], kwargs: FuncDependency) -> None:
        dep = args[0]
        rust_abi = self.package.abi_resolve_default(kwargs['rust_abi'])
        depname = self.package.get_dependency_name(rust_abi)

        if rust_abi == 'proc-macro':
            state.override_dependency(depname, dep, for_machine=MachineChoice.HOST)
            state.override_dependency(depname, dep, static=False, for_machine=MachineChoice.HOST)
            if state.environment.is_cross_build():
                state.override_dependency(depname, dep, for_machine=MachineChoice.BUILD)
                state.override_dependency(depname, dep, static=False, for_machine=MachineChoice.BUILD)
            return

        state.override_dependency(depname, dep)
        if self.package.abi_has_static(rust_abi):
            state.override_dependency(depname, dep, static=True)
        if self.package.abi_has_shared(rust_abi):
            state.override_dependency(depname, dep, static=False)

    @typed_pos_args('package.library', optargs=[(str, StructuredSources), StructuredSources])
    @typed_kwargs(
        'package.library',
        *LIBRARY_KWS_NO_NATIVE,
        DEPENDENCIES_KW,
        LINK_WITH_KW,
        LINK_WHOLE_KW,
        _BASE_LANG_KW.evolve(name='rust_args'),
    )
    def library_method(self, state: ModuleState, args: T.Tuple[
            T.Optional[T.Union[str, StructuredSources]],
            T.Optional[StructuredSources]], kwargs: RustPackageLibrary) -> T.Union[BothLibraries, SharedLibrary, StaticLibrary]:
        if not self.package.manifest.lib:
            raise MesonException("no [lib] section in Cargo package")
        if kwargs['rust_crate_type'] is not None:
            static = kwargs['rust_crate_type'] in {'lib', 'rlib', 'staticlib'}
            shared = kwargs['rust_crate_type'] in {'dylib', 'cdylib', 'proc-macro'}
        else:
            rust_abi = self.package.abi_resolve_default(kwargs['rust_abi'])
            static = self.package.abi_has_static(rust_abi)
            shared = self.package.abi_has_shared(rust_abi)
            if rust_abi == 'proc-macro':
                kwargs['rust_crate_type'] = 'proc-macro'
                kwargs['rust_abi'] = None
            else:
                kwargs['rust_abi'] = rust_abi
        return self._library_method(state, args, kwargs, static=static, shared=shared)

    @typed_pos_args('package.proc_macro', optargs=[(str, StructuredSources), StructuredSources])
    @typed_kwargs(
        'package.proc_macro',
        *SHARED_LIB_KWS_NO_NATIVE,
        DEPENDENCIES_KW,
        LINK_WITH_KW,
        LINK_WHOLE_KW,
        _BASE_LANG_KW.evolve(name='rust_args'),
    )
    def proc_macro_method(self, state: 'ModuleState', args: T.Tuple[
            T.Optional[T.Union[str, StructuredSources]],
            T.Optional[StructuredSources]], kwargs: RustPackageLibrary) -> SharedLibrary:
        if not self.package.manifest.lib:
            raise MesonException("no [lib] section in Cargo package")
        if 'proc-macro' not in self.package.manifest.lib.crate_type:
            raise MesonException("not a procedural macro crate")
        return self._proc_macro_method(state, args, kwargs)

    @typed_pos_args('package.shared_module', optargs=[(str, StructuredSources), StructuredSources])
    @typed_kwargs(
        'package.shared_module',
        *SHARED_MOD_KWS_NO_NATIVE,
        DEPENDENCIES_KW,
        LINK_WITH_KW,
        LINK_WHOLE_KW,
        _BASE_LANG_KW.evolve(name='rust_args'),
    )
    def shared_module_method(self, state: 'ModuleState', args: T.Tuple[
            T.Optional[T.Union[str, StructuredSources]],
            T.Optional[StructuredSources]], kwargs: RustPackageLibrary) -> SharedModule:
        if not self.package.manifest.lib:
            raise MesonException("no [lib] section in Cargo package")
        if 'cdylib' not in self.package.manifest.lib.crate_type:
            raise MesonException("not a cdylib crate")

        kwargs['rust_abi'] = None
        kwargs['rust_crate_type'] = 'cdylib'
        result = self._library_method(state, args, kwargs, shared=True, static=False, shared_mod=True)
        return T.cast('SharedModule', result)

    @typed_pos_args('package.executable', optargs=[(str, StructuredSources), StructuredSources])
    @typed_kwargs(
        'package.executable',
        *EXECUTABLE_KWS_NO_NATIVE,
        DEPENDENCIES_KW,
        LINK_WITH_KW,
        LINK_WHOLE_KW,
        _BASE_LANG_KW.evolve(name='rust_args'),
    )
    def executable_method(self, state: 'ModuleState', args: T.Tuple[
            T.Optional[T.Union[str, StructuredSources]],
            T.Optional[StructuredSources]], kwargs: RustPackageExecutable) -> Executable:
        """Builds executable targets from workspace bins."""
        tgt_args = self.validate_pos_args('package.executable', args)
        if not self.package.manifest.bin:
            raise MesonException("no [[bin]] section in Cargo package")

        sources: T.Union[StructuredSources, str]
        tgt_name, sources = tgt_args
        # If there's more than one binary, the first argument must be specified
        # and must be one of the keys in pkg.bin
        if not tgt_name:
            if len(self.package.manifest.bin) > 1:
                raise MesonException("Package has multiple binaries, you must specify which one to build as the first argument")
            # Single binary, use it
            tgt_name = next(iter(self.package.manifest.bin.keys()))
        else:
            if tgt_name not in self.package.manifest.bin:
                raise MesonException(f"Binary '{tgt_name}' not found.")

        if not sources:
            sources = os.path.relpath(os.path.join(self.package.path, self.package.manifest.bin[tgt_name].path),
                                      state.subdir)

        exe_args: T.Tuple[str, SourcesVarargsType] = (tgt_name, [sources])
        self.merge_kw_args(state, kwargs)
        return state._interpreter.build_target(state.current_node, exe_args, kwargs, Executable)


class RustSubproject(RustCrate):
    """Represents a Cargo subproject."""

    def __init__(self, rust_ws: RustWorkspace, package: cargo.PackageState) -> None:
        super().__init__(rust_ws, package)
        self.methods.update({
            'dependency': self.dependency_method,
        })

    @noPosargs
    @typed_kwargs('package.dependency',
                  KwargInfo('rust_abi', (str, NoneType), default=None, validator=in_set_validator({'rust', 'c', 'proc-macro'})))
    def dependency_method(self, state: ModuleState, args: T.List, kwargs: FuncDependency) -> Dependency:
        """Returns dependency for the package with the given ABI."""
        depname = self.package.get_dependency_name(kwargs['rust_abi'])
        return state.overridden_dependency(depname)


class RustModule(ExtensionModule):

    """A module that holds helper functions for rust."""

    INFO = ModuleInfo('rust', '0.57.0', stabilized='1.0.0')
    _bindgen_rust_target: T.Optional[str]
    rustdoc: PerMachine[T.Optional[ExternalProgram]] = PerMachine(None, None)

    def __init__(self, interpreter: Interpreter) -> None:
        super().__init__(interpreter)
        self._bindgen_bin: T.Optional[Program] = None
        if 'rust' in interpreter.compilers.host:
            rustc = T.cast('RustCompiler', interpreter.compilers.host['rust'])
            self._bindgen_rust_target = 'nightly' if rustc.is_nightly else rustc.version
        else:
            self._bindgen_rust_target = None
        self._bindgen_set_std = False
        self.methods.update({
            'test': self.test,
            'doctest': self.doctest,
            'bindgen': self.bindgen,
            'compiler_target': self.compiler_target,
            'proc_macro': self.proc_macro,
            'to_system_dependency': self.to_system_dependency,
            'workspace': self.workspace,
        })

    def test_common(self, funcname: str, state: ModuleState, args: T.Tuple[str, BuildTarget], kwargs: FuncRustTest) -> T.Tuple[Executable, _kwargs.FuncTest]:
        """Generate a rust test target from a given rust target.

        Rust puts its unitests inside its main source files, unlike most
        languages that put them in external files. This means that normally
        you have to define two separate targets with basically the same
        arguments to get tests:

        ```meson
        rust_lib_sources = [...]
        rust_lib = static_library(
            'rust_lib',
            rust_lib_sources,
        )

        rust_lib_test = executable(
            'rust_lib_test',
            rust_lib_sources,
            rust_args : ['--test'],
        )

        test(
            'rust_lib_test',
            rust_lib_test,
            protocol : 'rust',
        )
        ```

        This is all fine, but not very DRY. This method makes it much easier
        to define rust tests:

        ```meson
        rust = import('unstable-rust')

        rust_lib = static_library(
            'rust_lib',
            [sources],
        )

        rust.test('rust_lib_test', rust_lib)
        ```
        """
        if any(isinstance(t, Jar) for t in kwargs.get('link_with', [])):
            raise InvalidArguments('Rust tests cannot link with Jar targets')
        if any(isinstance(t, Jar) for t in kwargs.get('link_whole', [])):
            raise InvalidArguments('Rust tests cannot link with Jar targets')

        name = args[0]
        base_target: BuildTarget = args[1]
        if not base_target.uses_rust():
            raise InterpreterException(f'Second positional argument to rustmod.{funcname}() must be a rust based target')
        extra_args = kwargs['args']

        # Delete any arguments we don't want passed
        if '--test' in extra_args:
            mlog.warning(f'Do not add --test to rustmod.{funcname}() arguments')
            extra_args.remove('--test')
        if '--format' in extra_args:
            mlog.warning(f'Do not add --format to rustmod.{funcname}() arguments')
            i = extra_args.index('--format')
            # Also delete the argument to --format
            del extra_args[i + 1]
            del extra_args[i]
        for i, a in enumerate(extra_args):
            if isinstance(a, str) and a.startswith('--format='):
                del extra_args[i]
                break

        # We need to cast here, as currently these don't have protocol in them, but test itself does.
        tkwargs = T.cast('_kwargs.FuncTest', kwargs.copy())

        tkwargs['args'] = extra_args + ['--test', '--format', 'pretty']
        tkwargs['protocol'] = 'rust'

        new_target_kwargs = T.cast('ExecutableKeywordArguments', base_target.original_kwargs.copy())
        del new_target_kwargs['rust_crate_type']
        for kw in ('pic', 'prelink', 'rust_abi', 'version', 'soversion', 'darwin_versions', 'shortname'):
            if kw in new_target_kwargs:
                del new_target_kwargs[kw]  # type: ignore[misc]

        new_target_kwargs['install'] = False
        new_target_kwargs['dependencies'] = new_target_kwargs.get('dependencies', []) + kwargs['dependencies']
        new_target_kwargs['link_with'] = new_target_kwargs.get('link_with', []) + T.cast('T.List[BuildTargetTypes]', kwargs['link_with'])
        new_target_kwargs['link_whole'] = new_target_kwargs.get('link_whole', []) + kwargs['link_whole']

        lang_args = base_target.extra_args.copy()
        lang_args['rust'] = base_target.extra_args['rust'] + kwargs['rust_args'] + ['--test']
        new_target_kwargs['language_args'] = lang_args

        sources = T.cast('T.List[SourceOutputs]', base_target.sources.copy())
        sources.extend(base_target.generated)

        new_target = Executable(
            name, base_target.subdir, state.subproject, base_target.for_machine,
            sources, base_target.structured_sources,
            base_target.objects, base_target.environment, base_target.compilers,
            new_target_kwargs)
        return new_target, tkwargs

    @typed_pos_args('rust.test', str, BuildTarget)
    @typed_kwargs(
        'rust.test',
        *TEST_KWS,
        DEPENDENCIES_KW,
        LINK_WITH_KW.evolve(since='1.2.0'),
        LINK_WHOLE_KW.evolve(since='1.8.0'),
        *RUST_TEST_KWS,
    )
    def test(self, state: ModuleState, args: T.Tuple[str, BuildTarget], kwargs: FuncTest) -> ModuleReturnValue:
        name, _ = args
        new_target, tkwargs = self.test_common('test', state, args, kwargs)
        test: Test = self.interpreter.make_test(
            self.interpreter.current_node, (name, new_target), tkwargs)

        return ModuleReturnValue(None, [new_target, test])

    @FeatureNew('rust.doctest', '1.8.0')
    @typed_pos_args('rust.doctest', str, BuildTarget)
    @typed_kwargs(
        'rust.doctest',
        *TEST_KWS_NO_ARGS,
        DEPENDENCIES_KW,
        LINK_WITH_KW,
        LINK_WHOLE_KW,
        *RUST_TEST_KWS,
        KwargInfo(
            'args',
            ContainerTypeInfo(list, str),
            listify=True,
            default=[],
            validator=no_spaces_validator,
        ),
    )
    def doctest(self, state: ModuleState, args: T.Tuple[str, T.Union[SharedLibrary, StaticLibrary]], kwargs: FuncDoctest) -> ModuleReturnValue:
        name, base_target = args

        if not base_target.uses_rust():
            raise MesonException('doc tests are only supported for Rust targets')
        if not base_target.uses_rust_abi():
            raise MesonException("doc tests are not supported for rust_abi: 'c'")
        if state.environment.is_cross_build() and state.environment.need_exe_wrapper(base_target.for_machine):
            mlog.notice('skipping Rust doctests due to cross compilation', once=True)
            return ModuleReturnValue(None, [])

        # Link the base target's crate into the tests
        kwargs['link_with'].append(base_target)
        kwargs['depends'].append(base_target)
        workdir = kwargs['workdir']
        kwargs['workdir'] = None
        new_target, tkwargs = self.test_common('doctest', state, args, kwargs)

        # added automatically by rustdoc; keep things simple
        tkwargs['args'].remove('--test')

        # --test-args= is "parsed" simply via the Rust function split_whitespace().
        # This means no quoting nightmares (pfew) but it also means no spaces.
        # Unfortunately it's pretty hard at this point to accept e.g. CustomTarget,
        # because their paths may not be known.  This is not a big deal because the
        # user does not control the test harness, so make things easy and allow
        # strings only.
        if tkwargs['args']:
            tkwargs['args'] = ['--test-args=' + ' '.join(T.cast('T.Sequence[str]', tkwargs['args']))]
        if workdir:
            tkwargs['args'].append('--test-run-directory=' + workdir)

        if self.rustdoc[base_target.for_machine] is None:
            rustc = T.cast('RustCompiler', base_target.compilers['rust'])
            rustdoc = rustc.get_rustdoc()
            if rustdoc:
                self.rustdoc[base_target.for_machine] = ExternalProgram(rustdoc.get_exe())
            else:
                self.rustdoc[base_target.for_machine] = NonExistingExternalProgram()

        rustdoc_prog = self.rustdoc[base_target.for_machine]
        if not rustdoc_prog.found():
            raise MesonException(f'could not find rustdoc for {base_target.for_machine} machine')

        doctests: Doctest = self.interpreter.make_test(
            self.interpreter.current_node, (name, rustdoc_prog), tkwargs, Doctest)

        # Note that the new_target is intentionally not returned, as it
        # is only reached via the base_target and never built by "ninja",
        # so we need to complete its initialization here
        new_target.process_compilers_late()
        doctests.target = new_target
        base_target.doctests = doctests
        return ModuleReturnValue(None, [doctests])

    @noPosargs
    @typed_kwargs(
        'rust.bindgen',
        KwargInfo('c_args', ContainerTypeInfo(list, str), default=[], listify=True),
        KwargInfo('args', ContainerTypeInfo(list, str), default=[], listify=True),
        KwargInfo(
            'input',
            ContainerTypeInfo(list, (File, GeneratedList, BuildTarget, BothLibraries, ExtractedObjects, CustomTargetIndex, CustomTarget, str), allow_empty=False),
            default=[],
            listify=True,
            required=True,
        ),
        KwargInfo('language', (str, NoneType), since='1.4.0', validator=in_set_validator({'c', 'cpp'})),
        KwargInfo('bindgen_version', ContainerTypeInfo(list, str), default=[], listify=True, since='1.4.0'),
        INCLUDE_DIRECTORIES.evolve(since_values={ContainerTypeInfo(list, str): '1.0.0'}),
        OUTPUT_KW,
        KwargInfo(
            'output_inline_wrapper',
            str,
            default='',
            since='1.4.0',
        ),
        DEPENDENCIES_KW.evolve(since='1.0.0'),
    )
    def bindgen(self, state: ModuleState, args: T.List, kwargs: FuncBindgen) -> ModuleReturnValue:
        """Wrapper around bindgen to simplify its use.

        The main thing this simplifies is the use of `include_directory`
        objects, instead of having to pass a plethora of `-I` arguments.
        """
        header, *_deps = self.interpreter.source_strings_to_files(kwargs['input'])

        # Split File and Target dependencies to add pass to CustomTarget
        depends: T.List[SourceOutputs] = []
        depend_files: T.List[File] = []
        for d in _deps:
            if isinstance(d, File):
                depend_files.append(d)
            else:
                depends.append(d)

        # Copy to avoid subsequent calls mutating the original
        # TODO: if we want this to be per-machine we'll need a native kwarg
        clang_args = state.environment.properties.host.get_bindgen_clang_args().copy()

        # Look for --target in the rust command itself if there isn't one passed in clang_args
        target_arg = parse_target(clang_args)
        if not target_arg and 'rust' in state._interpreter.compilers.host:
            rust_args = state._interpreter.compilers.host['rust'].get_exe_args()
            target_arg = parse_target(rust_args)
            if target_arg:
                clang_args.append(f'--target={target_arg}')

        # Find the first C'ish compiler to fetch the default compiler flags
        # from. Append those to the bindgen flags to ensure we use a compatible
        # environment.
        comp = mesonlib.first(
            # The cast here shouldn't be necessary
            [state.environment.coredata.compilers.host.get(T.cast('Language', l))
             for l in ('c', 'cpp', 'objc', 'objcpp')],
            lambda x: x is not None,
        )
        if comp:
            clang_args.extend(comp.get_always_args())
        else:
            mlog.warning(textwrap.dedent('''\
                Using `rust.bindgen` without configuring C (or a C-like)
                language in Meson will skip compiler detection and can cause
                ABI incompatibilities due to missing crucial compiler flags.
                Consider calling `add_languages('c')` in your Meson build
                files.
            '''))

        for i in state.process_include_dirs(kwargs['include_directories']):
            # bindgen always uses clang, so it's safe to hardcode -I here
            clang_args.extend([f'-I{x}' for x in i.abs_string_list(
                state.environment.get_source_dir(), state.environment.get_build_dir())])
        if are_asserts_disabled_for_subproject(state.subproject, state.environment):
            clang_args.append('-DNDEBUG')

        for de in kwargs['dependencies']:
            for i in de.get_include_dirs():
                clang_args.extend([f'-I{x}' for x in i.abs_string_list(
                    state.environment.get_source_dir(), state.environment.get_build_dir())])
            clang_args.extend(de.get_all_compile_args())
            for s in de.get_sources():
                if isinstance(s, File):
                    depend_files.append(s)
                elif isinstance(s, CustomTarget):
                    depends.append(s)

        if self._bindgen_bin is None:
            self._bindgen_bin = state.find_program('bindgen', wanted=kwargs['bindgen_version'])
            if self._bindgen_rust_target is not None:
                _, _, err = mesonlib.Popen_safe(self._bindgen_bin.get_command() + ['--rust-target', self._bindgen_rust_target])
                # < 0.71: Sometimes this is "invalid Rust target" and
                # sometimes "invalid # rust target"
                # >= 0.71: error: invalid value '...' for '--rust-target <RUST_TARGET>': "..." is not a valid Rust target, accepted values are of the form ...
                # It's also much harder to hit this in 0.71 than in previous versions
                if 'Got an invalid' in err or 'is not a valid Rust target' in err:
                    self._bindgen_rust_target = None

            self._bindgen_set_std = mesonlib.version_compare(self._bindgen_bin.get_version(), '>= 0.71')

        name: str
        if isinstance(header, File):
            name = header.fname
        elif isinstance(header, (BuildTarget, BothLibraries, ExtractedObjects, StructuredSources)):
            raise InterpreterException('bindgen source file must be a C header, not an object or build target')
        else:
            name = header.get_outputs()[0]

        # bindgen assumes that C++ headers will be called .hpp. We want to
        # ensure that anything Meson considers a C++ header is treated as one.
        language = kwargs['language']
        if language is None:
            ext = os.path.splitext(name)[1][1:]
            if ext in lang_suffixes['cpp']:
                language = 'cpp'
            elif ext == 'h':
                language = 'c'
            else:
                raise InterpreterException(f'Unknown file type extension for: {name}')

        # We only want include directories and defines, other things may not be valid
        cargs = state.get_option(f'{language}_args', state.subproject)
        assert isinstance(cargs, list), 'for mypy'
        for a in itertools.chain(state.global_args.get(language, []), state.project_args.get(language, []), cargs):
            if a.startswith(('-I', '/I', '-D', '/D', '-U', '/U')):
                clang_args.append(a)

        if language == 'cpp':
            clang_args.extend(['-x', 'c++'])

        # Add the C++ standard to the clang arguments. Attempt to translate VS
        # extension versions into the nearest standard version
        std = state.get_option(f'{language}_std')
        assert isinstance(std, str), 'for mypy'
        if std.startswith('vc++'):
            if std.endswith('latest'):
                mlog.warning('Attempting to translate vc++latest into a clang compatible version.',
                             'Currently this is hardcoded for c++20', once=True, fatal=False)
                std = 'c++20'
            else:
                mlog.debug('The current C++ standard is a Visual Studio extension version.',
                           'bindgen will use a the nearest C++ standard instead')
                std = std[1:]

        if std != 'none':
            clang_args.append(f'-std={std}')

        inline_wrapper_args: T.List[str] = []
        outputs = [kwargs['output']]
        if kwargs['output_inline_wrapper']:
            # Todo drop this isinstance once Executable supports version_compare
            if isinstance(self._bindgen_bin, ExternalProgram):
                if mesonlib.version_compare(self._bindgen_bin.get_version(), '< 0.65'):
                    raise InterpreterException('\'output_inline_wrapper\' parameter of rust.bindgen requires bindgen-0.65 or newer')

            outputs.append(kwargs['output_inline_wrapper'])
            inline_wrapper_args = [
                '--experimental', '--wrap-static-fns',
                '--wrap-static-fns-path', os.path.join(state.environment.build_dir, '@OUTPUT1@')
            ]

        cmd = self._bindgen_bin.get_command() + \
            [
                '@INPUT@', '--output',
                os.path.join(state.environment.build_dir, '@OUTPUT0@')
            ] + \
            kwargs['args'] + inline_wrapper_args
        if self._bindgen_rust_target and '--rust-target' not in cmd:
            cmd.extend(['--rust-target', self._bindgen_rust_target])
        if self._bindgen_set_std and '--rust-edition' not in cmd:
            try:
                rust_std = state.environment.coredata.optstore.get_value_for('rust_std')
            except KeyError:
                rust_std = 'none'
            assert isinstance(rust_std, str), 'for mypy'
            if rust_std != 'none':
                cmd.extend(['--rust-edition', rust_std])
        cmd.append('--')
        cmd.extend(kwargs['c_args'])
        cmd.extend(clang_args)
        cmd.extend(['-MD', '-MQ', '@INPUT@', '-MF', '@DEPFILE@'])

        target = CustomTarget(
            f'rustmod-bindgen-{name}'.replace('/', '_'),
            state.subdir,
            state.subproject,
            state.environment,
            cmd,
            [header],
            outputs,
            depfile='@PLAINNAME@.d',
            extra_depends=depends,
            depend_files=depend_files,
            backend=state.backend,
            description='Generating bindings for Rust {}',
        )

        return ModuleReturnValue(target, [target])

    @FeatureNew('rust.compiler_target', '1.11.0')
    @noPosargs
    @typed_kwargs('rust.compiler_target', NATIVE_KW)
    def compiler_target(self, state: ModuleState, args: T.List, kwargs: '_kwargs.NativeKW') -> str:
        """Returns the Rust target triple for the specified machine's Rust compiler."""
        for_machine = kwargs['native']
        compilers = state._interpreter.coredata.compilers[for_machine]
        if 'rust' in compilers:
            rustc = T.cast('RustCompiler', compilers['rust'])
            return rustc.get_target_triple()
        else:
            raise MesonException(f'No Rust compiler was requested for the {for_machine} machine')

    # Allow a limited set of kwargs, but still use the full set of typed_kwargs()
    # because it could be setting required default values.
    @FeatureNew('rust.proc_macro', '1.3.0')
    @permittedKwargs({'rust_args', 'rust_dependency_map', 'sources', 'dependencies', 'extra_files',
                      'link_args', 'link_depends', 'link_with', 'override_options'})
    @typed_pos_args('rust.proc_macro', str, varargs=SOURCES_VARARGS)
    @typed_kwargs('rust.proc_macro', *SHARED_LIB_KWS)
    def proc_macro(self, state: ModuleState, args: T.Tuple[str, SourcesVarargsType], kwargs: _kwargs.SharedLibrary) -> SharedLibrary:
        kwargs['native'] = MachineChoice.BUILD
        kwargs['rust_crate_type'] = 'proc-macro'
        kwargs['rust_args'] = kwargs['rust_args'] + ['--extern', 'proc_macro']
        target = state._interpreter.build_target(state.current_node, args, kwargs, SharedLibrary)
        return target

    @FeatureNew('rust.to_system_dependency', '1.11.0')
    @typed_pos_args('rust.to_system_dependency', Dependency, optargs=[str])
    @noKwargs
    def to_system_dependency(self, state: ModuleState, args: T.Tuple[Dependency, T.Optional[str]], kwargs: TYPE_kwargs) -> Dependency:
        dep, depname = args
        return dep_to_system_dependency(dep, depname)

    @FeatureNew('rust.workspace', '1.11.0')
    @noPosargs
    @typed_kwargs(
        'rust.workspace',
        KwargInfo('default_features', (bool, NoneType), default=None),
        KwargInfo(
            'features',
            (ContainerTypeInfo(list, str), NoneType),
            default=None,
            listify=True,
        ),
    )
    def workspace(self, state: ModuleState, args: T.List, kwargs: FuncWorkspace) -> RustWorkspace:
        """Creates a Rust workspace object, controlling the build of
           all the packages in a Cargo.lock file."""
        if self.interpreter.cargo is None:
            raise MesonException("rust.workspace() requires a Cargo project (Cargo.toml and Cargo.lock)")

        self.interpreter.add_languages(['rust'], True, MachineChoice.HOST)
        self.interpreter.add_languages(['rust'], True, MachineChoice.BUILD)

        default_features = kwargs['default_features']
        features = kwargs['features']
        if default_features is not None or features is not None:
            # If custom features are provided, default_features = None should be treated as True
            if default_features is None:
                default_features = True

            cargo_features = ['default'] if default_features else []
            if features is not None:
                cargo_features.extend(features)
            self.interpreter.cargo.features = cargo_features

        ws = self.interpreter.cargo.load_workspace(state.root_subdir)

        # Cargo projects may not have a subprojects directory, because
        # dependencies are declared in Cargo.toml rather than .wrap files.
        # Create it now so that the wrap resolver can download the crates there
        os.makedirs(self.interpreter.environment.wrap_resolver.subdir_root, exist_ok=True)

        return RustWorkspace(self.interpreter, ws)


def initialize(interp: Interpreter) -> RustModule:
    return RustModule(interp)



#! /usr/bin/env python3

import json
import sys
import os

cc = None
output = None

# Only the ninja backend produces compile_commands.json
if sys.argv[1] == 'ninja':
    with open('compile_commands.json') as f:
        cc = json.load(f)
    output = {x['output'] for x in cc}

for obj in sys.argv[2:]:
    if not os.path.exists(obj):
        sys.exit(f'File {obj} not found.')
    if sys.argv[1] == 'ninja' and obj not in output:
        sys.exit(1)
    print('Verified', obj)



# SPDX-License-Identifier: Apache-2.0

"""Wrapper for rc.exe to make it output included files."""

from __future__ import annotations

import argparse
import subprocess
import shutil
import typing as T


parser = argparse.ArgumentParser()
parser.add_argument('--rc')
parser.add_argument('--cl')


def run(args: T.List[str]) -> int:
    options, rc_args = parser.parse_known_args(args)
    target = rc_args[-1] if rc_args else None

    rc = options.rc if options.rc else shutil.which('rc.exe')
    if rc is None:
        print('rc.exe not found')
        return 1

    cl = options.cl if options.cl else shutil.which('cl.exe')
    if cl is None:
        print('cl.exe not found')
        return 1

    # Use preprocessor to display include files
    include_args = [a for a in rc_args if a.startswith(('/I', '-I'))]
    cmd = [cl, '/showIncludes', '/EP', '/nologo', '/DRC_INVOKED'] + include_args + ['/Tc' + target]
    result = subprocess.call(cmd, stdout=subprocess.DEVNULL)
    if result != 0:
        print('Error running preprocessor to find resource dependencies')
        # continue anyway. rc.exe should catch the error later

    cmd = [rc] + rc_args
    return subprocess.call(cmd)



#!/usr/bin/env python3

import json
import argparse
import stat
import textwrap
import shutil
import subprocess
from tempfile import TemporaryDirectory
from pathlib import Path
import typing as T

image_namespace = 'mesonbuild'

image_def_file = 'image.json'
install_script = 'install.sh'

class ImageDef:
    def __init__(self, image_dir: Path) -> None:
        path = image_dir / image_def_file
        data = json.loads(path.read_text(encoding='utf-8'))

        assert isinstance(data, dict)
        assert all([x in data for x in ['base_image', 'env']])
        assert isinstance(data['base_image'], str)
        assert isinstance(data['env'],  dict)

        self.base_image: str = data['base_image']
        self.args: T.List[str] = data.get('args', [])
        self.env: T.Dict[str, str] = data['env']

class BuilderBase():
    def __init__(self, data_dir: Path, temp_dir: Path) -> None:
        self.data_dir = data_dir
        self.temp_dir = temp_dir

        self.common_sh = self.data_dir.parent / 'common.sh'
        self.common_sh = self.common_sh.resolve(strict=True)
        self.validate_data_dir()

        self.image_def = ImageDef(self.data_dir)

        self.docker = shutil.which('docker')
        self.git = shutil.which('git')
        if self.docker is None:
            raise RuntimeError('Unable to find docker')
        if self.git is None:
            raise RuntimeError('Unable to find git')

    def validate_data_dir(self) -> None:
        files = [
            self.data_dir / image_def_file,
            self.data_dir / install_script,
        ]
        if not self.data_dir.exists():
            raise RuntimeError(f'{self.data_dir.as_posix()} does not exist')
        for i in files:
            if not i.exists():
                raise RuntimeError(f'{i.as_posix()} does not exist')
            if not i.is_file():
                raise RuntimeError(f'{i.as_posix()} is not a regular file')

class Builder(BuilderBase):
    def gen_bashrc(self) -> None:
        out_file = self.temp_dir / 'env_vars.sh'
        out_data = ''

        # run_tests.py parameters
        self.image_def.env['CI_ARGS'] = ' '.join(self.image_def.args)

        for key, val in self.image_def.env.items():
            out_data += f'export {key}="{val}"\n'

        # Also add /ci to PATH
        out_data += 'export PATH="/ci:$PATH"\n'

        out_data += '''
            if [ -f "$HOME/.cargo/env" ]; then
                source "$HOME/.cargo/env"
            fi
        '''

        out_file.write_text(out_data, encoding='utf-8')

        # make it executable
        mode = out_file.stat().st_mode
        out_file.chmod(mode | stat.S_IEXEC)

    def gen_dockerfile(self) -> None:
        out_file = self.temp_dir / 'Dockerfile'
        out_data = textwrap.dedent(f'''\
            FROM {self.image_def.base_image}

            ADD install.sh  /ci/install.sh
            ADD common.sh   /ci/common.sh
            ADD env_vars.sh /ci/env_vars.sh
            RUN /ci/install.sh
        ''')

        out_file.write_text(out_data, encoding='utf-8')

    def do_build(self) -> None:
        # copy files
        for i in self.data_dir.iterdir():
            shutil.copy(str(i), str(self.temp_dir))
        shutil.copy(str(self.common_sh), str(self.temp_dir))

        self.gen_bashrc()
        self.gen_dockerfile()

        cmd_git = [self.git, 'rev-parse', '--short', 'HEAD']
        res = subprocess.run(cmd_git, cwd=self.data_dir, stdout=subprocess.PIPE)
        if res.returncode != 0:
            raise RuntimeError('Failed to get the current commit hash')
        commit_hash = res.stdout.decode().strip()

        cmd = [
            self.docker, 'build',
            '-t', f'{image_namespace}/{self.data_dir.name}:latest',
            '-t', f'{image_namespace}/{self.data_dir.name}:{commit_hash}',
            '--pull',
            self.temp_dir.as_posix(),
        ]
        if subprocess.run(cmd).returncode != 0:
            raise RuntimeError('Failed to build the docker image')

class ImageTester(BuilderBase):
    def __init__(self, data_dir: Path, temp_dir: Path, ci_root: Path) -> None:
        super().__init__(data_dir, temp_dir)
        self.meson_root = ci_root.parent.parent.resolve()

    def gen_dockerfile(self) -> None:
        out_file = self.temp_dir / 'Dockerfile'
        out_data = textwrap.dedent(f'''\
            FROM {image_namespace}/{self.data_dir.name}

            ADD meson /meson
        ''')

        out_file.write_text(out_data, encoding='utf-8')

    def copy_meson(self) -> None:
        shutil.copytree(
            self.meson_root,
            self.temp_dir / 'meson',
            symlinks=True,
            ignore=shutil.ignore_patterns(
                '.git',
                '*_cache',
                '__pycache__',
                # 'work area',
                self.temp_dir.name,
            ),
        )

    def do_test(self, tty: bool = False) -> None:
        self.copy_meson()
        self.gen_dockerfile()

        try:
            build_cmd = [
                self.docker, 'build',
                '-t', 'meson_test_image',
                self.temp_dir.as_posix(),
            ]
            if subprocess.run(build_cmd).returncode != 0:
                raise RuntimeError('Failed to build the test docker image')

            test_cmd = []
            if tty:
                test_cmd = [
                    self.docker, 'run', '--rm', '-t', '-i', 'meson_test_image',
                    '/bin/bash', '-c', ''
                    + 'cd meson;'
                    + 'source /ci/env_vars.sh;'
                    + f'echo -e "\\n\\nInteractive test shell in the {image_namespace}/{self.data_dir.name} container with the current meson tree";'
                    + 'echo -e "The file ci/ciimage/user.sh will be sourced if it exists to enable user specific configurations";'
                    + 'echo -e "Run the following command to run all CI tests: ./run_tests.py $CI_ARGS\\n\\n";'
                    + '[ -f ci/ciimage/user.sh ] && exec /bin/bash --init-file ci/ciimage/user.sh;'
                    + 'exec /bin/bash;'
                ]
            else:
                test_cmd = [
                    self.docker, 'run', '--rm', '-t', 'meson_test_image',
                    '/bin/bash', '-xc', 'source /ci/env_vars.sh; cd meson; ./run_tests.py $CI_ARGS'
                ]

            if subprocess.run(test_cmd).returncode != 0 and not tty:
                raise RuntimeError('Running tests failed')
        finally:
            cleanup_cmd = [self.docker, 'rmi', '-f', 'meson_test_image']
            subprocess.run(cleanup_cmd).returncode

class ImageTTY(BuilderBase):
    def __init__(self, data_dir: Path, temp_dir: Path, ci_root: Path) -> None:
        super().__init__(data_dir, temp_dir)
        self.meson_root = ci_root.parent.parent.resolve()

    def do_run(self) -> None:
        try:
            tty_cmd = [
                self.docker, 'run',
                '--name', 'meson_test_container', '-t', '-i', '-v', f'{self.meson_root.as_posix()}:/meson',
                f'{image_namespace}/{self.data_dir.name}',
                '/bin/bash', '-c', ''
                    + 'cd meson;'
                    + 'source /ci/env_vars.sh;'
                    + f'echo -e "\\n\\nInteractive test shell in the {image_namespace}/{self.data_dir.name} container with the current meson tree";'
                    + 'echo -e "The file ci/ciimage/user.sh will be sourced if it exists to enable user specific configurations";'
                    + 'echo -e "Run the following command to run all CI tests: ./run_tests.py $CI_ARGS\\n\\n";'
                    + '[ -f ci/ciimage/user.sh ] && exec /bin/bash --init-file ci/ciimage/user.sh;'
                    + 'exec /bin/bash;'
            ]
            subprocess.run(tty_cmd).returncode != 0
        finally:
            cleanup_cmd = [self.docker, 'rm', '-f', 'meson_test_container']
            subprocess.run(cleanup_cmd).returncode


def main() -> None:
    parser = argparse.ArgumentParser(description='Meson CI image builder')
    parser.add_argument('what', type=str, help='Which image to build / test')
    parser.add_argument('-t', '--type', choices=['build', 'test', 'testTTY', 'TTY'], help='What to do', required=True)

    args = parser.parse_args()

    ci_root = Path(__file__).parent
    ci_data = ci_root / args.what

    with TemporaryDirectory(prefix=f'{args.type}_{args.what}_', dir=ci_root) as td:
        ci_build = Path(td)
        print(f'Build dir: {ci_build}')

        if args.type == 'build':
            builder = Builder(ci_data, ci_build)
            builder.do_build()
        elif args.type == 'test':
            tester = ImageTester(ci_data, ci_build, ci_root)
            tester.do_test()
        elif args.type == 'testTTY':
            tester = ImageTester(ci_data, ci_build, ci_root)
            tester.do_test(tty=True)
        elif args.type == 'TTY':
            tester = ImageTTY(ci_data, ci_build, ci_root)
            tester.do_run()

if __name__ == '__main__':
    main()



# SPDX-License-Identifier: Apache-2.0
# Copyright 2021 The Meson development team
# Copyright © 2024-2025 Intel Corporation

from __future__ import annotations
import json
import os
import pickle
import subprocess
import tempfile
import subprocess
import textwrap
import shutil
from unittest import skipIf, SkipTest
from pathlib import Path

from .baseplatformtests import BasePlatformTests
from .helpers import *
from mesonbuild.mesonlib import EnvironmentVariables, ExecutableSerialisation, MesonException, is_linux, python_command, windows_proof_rmtree
from mesonbuild.mformat import Formatter, match_path
from mesonbuild.optinterpreter import OptionInterpreter, OptionException
from mesonbuild.options import OptionStore
from run_tests import Backend

@skipIf(IS_CI and not is_linux(), "Run only on fast platforms")
class PlatformAgnosticTests(BasePlatformTests):
    '''
    Tests that does not need to run on all platforms during CI
    '''

    def test_relative_find_program(self):
        '''
        Tests that find_program() with a relative path does not find the program
        in current workdir.
        '''
        testdir = os.path.join(self.unit_test_dir, '100 relative find program')
        self.init(testdir, workdir=testdir)

    def test_invalid_option_names(self):
        store = OptionStore(False)
        interp = OptionInterpreter(store, '')

        def write_file(code: str):
            with tempfile.NamedTemporaryFile('w', dir=self.builddir, encoding='utf-8', delete=False) as f:
                f.write(code)
                return f.name

        fname = write_file("option('default_library', type: 'string')")
        self.assertRaisesRegex(OptionException, 'Option name default_library is reserved.',
                               interp.process, fname)

        fname = write_file("option('c_anything', type: 'string')")
        self.assertRaisesRegex(OptionException, 'Option name c_anything is reserved.',
                               interp.process, fname)

        fname = write_file("option('b_anything', type: 'string')")
        self.assertRaisesRegex(OptionException, 'Option name b_anything is reserved.',
                               interp.process, fname)

        fname = write_file("option('backend_anything', type: 'string')")
        self.assertRaisesRegex(OptionException, 'Option name backend_anything is reserved.',
                               interp.process, fname)

        fname = write_file("option('foo.bar', type: 'string')")
        self.assertRaisesRegex(OptionException, 'Option names can only contain letters, numbers or dashes.',
                               interp.process, fname)

        # platlib is allowed, only python.platlib is reserved.
        fname = write_file("option('platlib', type: 'string')")
        interp.process(fname)

    def test_option_validation(self):
        """Test cases that are not catch by the optinterpreter itself."""
        store = OptionStore(False)
        interp = OptionInterpreter(store, '')

        def write_file(code: str):
            with tempfile.NamedTemporaryFile('w', dir=self.builddir, encoding='utf-8', delete=False) as f:
                f.write(code)
                return f.name

        fname = write_file("option('intminmax', type: 'integer', value: 10, min: 0, max: 5)")
        self.assertRaisesRegex(MesonException, 'Value 10 for option "intminmax" is more than maximum value 5.',
                               interp.process, fname)

        fname = write_file("option('array', type: 'array', choices : ['one', 'two', 'three'], value : ['one', 'four'])")
        self.assertRaisesRegex(MesonException, 'Value "four" for option "array" is not in allowed choices: "one, two, three"',
                               interp.process, fname)

        fname = write_file("option('array', type: 'array', choices : ['one', 'two', 'three'], value : ['four', 'five', 'six'])")
        self.assertRaisesRegex(MesonException, 'Values "four, five, six" for option "array" are not in allowed choices: "one, two, three"',
                               interp.process, fname)

    def test_python_dependency_without_pkgconfig(self):
        testdir = os.path.join(self.unit_test_dir, '102 python without pkgconfig')
        self.init(testdir, override_envvars={'PKG_CONFIG': 'notfound'})

    def test_vala_target_with_internal_glib(self):
        testdir = os.path.join(self.unit_test_dir, '131 vala internal glib')
        for run in [{ 'version': '2.84.4', 'expected': '2.84'}, { 'version': '2.85.2', 'expected': '2.84' }]:
            self.new_builddir()
            self.init(testdir, extra_args=[f'-Dglib-version={run["version"]}'])
            try:
                with open(os.path.join(self.builddir, 'meson-info', 'intro-targets.json'), 'r', encoding='utf-8') as tgt_intro:
                    intro = json.load(tgt_intro)
                    target = list(filter(lambda tgt: tgt['name'] == 'vala-tgt', intro))
                    self.assertLength(target, 1)
                    sources = target[0]['target_sources']
                    vala_sources = filter(lambda src: src.get('language') == 'vala', sources)
                    for src in vala_sources:
                        self.assertIn(('--target-glib', run['expected']), zip(src['parameters'], src['parameters'][1:]))
            except FileNotFoundError:
                self.skipTest('Current backend does not produce introspection data')

    def test_debug_function_outputs_to_meson_log(self):
        testdir = os.path.join(self.unit_test_dir, '104 debug function')
        log_msg = 'This is an example debug output, should only end up in debug log'
        output = self.init(testdir)

        # Check if message is not printed to stdout while configuring
        self.assertNotIn(log_msg, output)

        # Check if message is written to the meson log
        mesonlog = self.get_meson_log_raw()
        self.assertIn(log_msg, mesonlog)

    def test_new_subproject_reconfigure(self):
        testdir = os.path.join(self.unit_test_dir, '108 new subproject on reconfigure')
        self.init(testdir)
        self.build()

        # Enable the subproject "foo" and reconfigure, this is used to fail
        # because per-subproject builtin options were not initialized:
        # https://github.com/mesonbuild/meson/issues/10225.
        self.setconf('-Dfoo=enabled')
        self.build('reconfigure')

    def check_connectivity(self):
        import urllib
        try:
            with urllib.request.urlopen('https://wrapdb.mesonbuild.com') as p:
                pass
        except urllib.error.URLError as e:
            self.skipTest('No internet connectivity: ' + str(e))

    def test_update_wrapdb(self):
        self.check_connectivity()
        # Write the project into a temporary directory because it will add files
        # into subprojects/ and we don't want to pollute meson source tree.
        with tempfile.TemporaryDirectory() as testdir:
            with Path(testdir, 'meson.build').open('w', encoding='utf-8') as f:
                f.write(textwrap.dedent(
                    '''
                    project('wrap update-db',
                      default_options: ['wrap_mode=forcefallback'])

                    zlib_dep = dependency('zlib')
                    assert(zlib_dep.type_name() == 'internal')
                    '''))
            subprocess.check_call(self.wrap_command + ['update-db'], cwd=testdir)
            self.init(testdir, workdir=testdir)

    def test_none_backend(self):
        testdir = os.path.join(self.python_test_dir, '7 install path')

        self.init(testdir, extra_args=['--backend=none'], override_envvars={'NINJA': 'absolutely false command'})
        self.assertPathDoesNotExist(os.path.join(self.builddir, 'build.ninja'))

        self.run_tests(inprocess=True, override_envvars={})

        out = self._run(self.meson_command + ['install', f'--destdir={self.installdir}'], workdir=self.builddir)
        self.assertNotIn('Only ninja backend is supported to rebuild the project before installation.', out)

        with open(os.path.join(testdir, 'test.json'), 'rb') as f:
            dat = json.load(f)
        for i in dat['installed']:
            self.assertPathExists(os.path.join(self.installdir, i['file']))

    def test_change_backend(self):
        if self.backend != Backend.ninja:
            raise SkipTest('Only useful to test if backend is ninja.')

        testdir = os.path.join(self.python_test_dir, '7 install path')
        self.init(testdir)

        # no-op change works
        with self.subTest('set the option to the same value'):
            self.setconf('--backend=ninja')
            self.init(testdir, extra_args=['--reconfigure', '--backend=ninja'])

        # Change backend option is not allowed
        with self.subTest('Changing the backend'):
            with self.assertRaises(subprocess.CalledProcessError) as cm:
                self.setconf('-Dbackend=none')
            self.assertIn('ERROR: Tried to modify read only option "backend"', cm.exception.stdout)

        # Check that the new value was not written in the store.
        with self.subTest('option is stored correctly'):
            self.assertEqual(self.getconf('backend'), 'ninja')

        # Wipe with a different backend is allowed
        with self.subTest('Changing the backend with wipe'):
            self.init(testdir, extra_args=['--wipe', '--backend=none'])

            self.assertEqual(self.getconf('backend'), 'none')

    def test_validate_dirs(self):
        testdir = os.path.join(self.common_test_dir, '1 trivial')

        # Using parent as source directory should fail
        self.builddir = os.path.dirname(os.getcwd())
        with self.assertRaises(subprocess.CalledProcessError) as cm:
            self.init(testdir)
        self.assertIn('cannot be a parent of source directory', cm.exception.stdout)

        # Reconfigure of empty builddir should work
        self.new_builddir()
        self.init(testdir, extra_args=['--reconfigure'])

        # Reconfigure of not empty builddir should work
        self.new_builddir()
        Path(self.builddir, 'dummy').touch()
        self.init(testdir, extra_args=['--reconfigure', '--buildtype=custom'])

        # Setup a valid builddir should update options but not reconfigure
        self.assertEqual(self.getconf('buildtype'), 'custom')
        o = self.init(testdir, extra_args=['-Dbuildtype=release'])
        self.assertIn('Directory already configured', o)
        self.assertNotIn('The Meson build system', o)
        self.assertEqual(self.getconf('buildtype'), 'release')

        # Wipe of empty builddir should work
        self.new_builddir()
        self.init(testdir, extra_args=['--wipe'])

        # Wipe of partial builddir should work
        self.new_builddir()
        Path(self.builddir, 'meson-private').mkdir()
        Path(self.builddir, 'dummy').touch()
        self.init(testdir, extra_args=['--wipe'])

        # Wipe of not empty builddir should fail
        self.new_builddir()
        Path(self.builddir, 'dummy').touch()
        with self.assertRaises(subprocess.CalledProcessError) as cm:
            self.init(testdir, extra_args=['--wipe'])
        self.assertIn('Directory is not empty', cm.exception.stdout)

    def test_scripts_loaded_modules(self):
        '''
        Simulate a wrapped command, as done for custom_target() that capture
        output. The script will print all python modules loaded and we verify
        that it contains only an acceptable subset. Loading too many modules
        slows down the build when many custom targets get wrapped.

        This list must not be edited without a clear rationale for why it is
        acceptable to do so!
        '''
        es = ExecutableSerialisation(python_command + ['-c', 'exit(0)'], env=EnvironmentVariables())
        p = Path(self.builddir, 'exe.dat')
        with p.open('wb') as f:
            pickle.dump(es, f)
        cmd = self.meson_command + ['--internal', 'test_loaded_modules', '--unpickle', str(p)]
        p = subprocess.run(cmd, stdout=subprocess.PIPE)
        all_modules = json.loads(p.stdout.splitlines()[0])
        meson_modules = [m for m in all_modules if m.startswith('mesonbuild')]
        expected_meson_modules = [
            'mesonbuild',
            'mesonbuild._pathlib',
            'mesonbuild.utils',
            'mesonbuild.utils.core',
            'mesonbuild.mesonmain',
            'mesonbuild.mlog',
            'mesonbuild.scripts',
            'mesonbuild.scripts.meson_exe',
            'mesonbuild.scripts.test_loaded_modules'
        ]
        self.assertEqual(sorted(expected_meson_modules), sorted(meson_modules))

    def test_setup_loaded_modules(self):
        '''
        Execute a very basic meson.build and capture a list of all python
        modules loaded. We verify that it contains only an acceptable subset.
        Loading too many modules slows down `meson setup` startup time and
        gives a perception that meson is slow.

        Adding more modules to the default startup flow is not an unreasonable
        thing to do as new features are added, but keeping track of them is
        good.
        '''
        testdir = os.path.join(self.unit_test_dir, '116 empty project')

        self.init(testdir)
        self._run(self.meson_command + ['--internal', 'regenerate', '--profile-self', testdir, self.builddir])
        with open(os.path.join(self.builddir, 'meson-logs', 'profile-startup-modules.json'), encoding='utf-8') as f:
                data = json.load(f)['meson']

        with open(os.path.join(testdir, 'expected_mods.json'), encoding='utf-8') as f:
            expected = json.load(f)['meson']

        self.assertEqual(data['modules'], expected['modules'])
        self.assertEqual(data['count'], expected['count'])

    def test_meson_package_cache_dir(self):
        # Copy testdir into temporary directory to not pollute meson source tree.
        testdir = os.path.join(self.unit_test_dir, '118 meson package cache dir')
        srcdir = os.path.join(self.builddir, 'srctree')
        shutil.copytree(testdir, srcdir)
        builddir = os.path.join(srcdir, '_build')
        self.change_builddir(builddir)
        self.init(srcdir, override_envvars={'MESON_PACKAGE_CACHE_DIR': os.path.join(srcdir, 'cache_dir')})

    def test_cmake_openssl_not_found_bug(self):
        """Issue #12098"""
        testdir = os.path.join(self.unit_test_dir, '119 openssl cmake bug')
        self.meson_native_files.append(os.path.join(testdir, 'nativefile.ini'))
        out = self.init(testdir, allow_fail=True)
        self.assertNotIn('Unhandled python exception', out)

    def test_editorconfig_match_path(self):
        '''match_path function used to parse editorconfig in meson format'''
        cases = [
            ('a.txt', '*.txt', True),
            ('a.txt', '?.txt', True),
            ('a.txt', 'a.t?t', True),
            ('a.txt', '*.build', False),

            ('/a.txt', '*.txt', True),
            ('/a.txt', '/*.txt', True),
            ('a.txt', '/*.txt', False),

            ('a/b/c.txt', 'a/b/*.txt', True),
            ('a/b/c.txt', 'a/*/*.txt', True),
            ('a/b/c.txt', '*/*.txt', True),
            ('a/b/c.txt', 'b/*.txt', True),
            ('a/b/c.txt', 'a/*.txt', False),

            ('a/b/c/d.txt', 'a/**/*.txt', True),
            ('a/b/c/d.txt', 'a/*', False),
            ('a/b/c/d.txt', 'a/**', True),

            ('a.txt', '[abc].txt', True),
            ('a.txt', '[!xyz].txt', True),
            ('a.txt', '[xyz].txt', False),
            ('a.txt', '[!abc].txt', False),

            ('a.txt', '{a,b,c}.txt', True),
            ('a.txt', '*.{txt,tex,cpp}', True),
            ('a.hpp', '*.{txt,tex,cpp}', False),

            ('a1.txt', 'a{0..9}.txt', True),
            ('a001.txt', 'a{0..9}.txt', True),
            ('a-1.txt', 'a{-10..10}.txt', True),
            ('a99.txt', 'a{0..9}.txt', False),
            ('a099.txt', 'a{0..9}.txt', False),
            ('a-1.txt', 'a{0..10}.txt', False),
        ]

        for filename, pattern, expected in cases:
            self.assertTrue(match_path(filename, pattern) is expected, f'{filename} -> {pattern}')

    def test_format_invalid_config_key(self) -> None:
        fd, fname = tempfile.mkstemp(suffix='.ini', text=True)
        self.addCleanup(os.unlink, fname)

        with os.fdopen(fd, 'w', encoding='utf-8') as handle:
            handle.write('not_an_option = 42\n')

        with self.assertRaises(MesonException):
            Formatter(Path(fname), use_editor_config=False, fetch_subdirs=False)

    def test_format_invalid_config_value(self) -> None:
        fd, fname = tempfile.mkstemp(suffix='.ini', text=True)
        self.addCleanup(os.unlink, fname)

        with os.fdopen(fd, 'w', encoding='utf-8') as handle:
            handle.write('max_line_length = string\n')

        with self.assertRaises(MesonException):
            Formatter(Path(fname), use_editor_config=False, fetch_subdirs=False)

    def test_format_invalid_editorconfig_value(self) -> None:
        dirpath = tempfile.mkdtemp()
        self.addCleanup(windows_proof_rmtree, dirpath)

        editorconfig = Path(dirpath, '.editorconfig')
        with open(editorconfig, 'w', encoding='utf-8') as handle:
            handle.write('[*]\n')
            handle.write('indent_size = string\n')

        formatter = Formatter(None, use_editor_config=True, fetch_subdirs=False)
        with self.assertRaises(MesonException):
            formatter.load_editor_config(editorconfig)

    def test_format_empty_file(self) -> None:
        formatter = Formatter(None, use_editor_config=False, fetch_subdirs=False)
        for code in ('', '\n'):
            formatted = formatter.format(code, Path())
            self.assertEqual('\n', formatted)

    def test_format_indent_comment_in_brackets(self) -> None:
        """Ensure comments in arrays and dicts are correctly indented"""
        formatter = Formatter(None, use_editor_config=False, fetch_subdirs=False)
        code = 'a = [\n    # comment\n]\n'
        formatted = formatter.format(code, Path())
        self.assertEqual(code, formatted)

        code = 'a = [\n    # comment\n    1,\n]\n'
        formatted = formatter.format(code, Path())
        self.assertEqual(code, formatted)

        code = 'a = {\n    # comment\n}\n'
        formatted = formatter.format(code, Path())
        self.assertEqual(code, formatted)

    def test_error_configuring_subdir(self):
        testdir = os.path.join(self.common_test_dir, '152 index customtarget')
        out = self.init(os.path.join(testdir, 'subdir'), allow_fail=True)

        self.assertIn('first statement must be a call to project()', out)
        # provide guidance diagnostics by finding a file whose first AST statement is project()
        self.assertIn(f'Did you mean to run meson from the directory: "{testdir}"?', out)

    def test_reconfigure_base_options(self):
        testdir = os.path.join(self.unit_test_dir, '123 reconfigure base options')
        out = self.init(testdir, extra_args=['-Db_ndebug=true'])
        self.assertIn('\nMessage: b_ndebug: true\n', out)
        self.assertIn('\nMessage: c_std: c89\n', out)

        out = self.init(testdir, extra_args=['--reconfigure', '-Db_ndebug=if-release', '-Dsub:b_ndebug=false', '-Dc_std=c99', '-Dsub:c_std=c11'])
        self.assertIn('\n    b_ndebug    : if-release\n', out)
        self.assertIn('\n    c_std       : c99\n', out)
        self.assertIn('\n    sub:b_ndebug: false\n', out)
        self.assertIn('\n    sub:c_std   : c11\n', out)

    def test_setup_with_unknown_option(self):
        testdir = os.path.join(self.common_test_dir, '1 trivial')

        with self.subTest('unknown user option'):
            out = self.init(testdir, extra_args=['-Dnot_an_option=1'], allow_fail=True)
            self.assertIn('ERROR: Unknown option: "not_an_option"', out)

        with self.subTest('unknown builtin option'):
            self.new_builddir()
            out = self.init(testdir, extra_args=['-Db_not_an_option=1'], allow_fail=True)
            self.assertIn('ERROR: Unknown option: "b_not_an_option"', out)


    def test_configure_new_option(self) -> None:
        """Adding a new option without reconfiguring should work."""
        testdir = self.copy_srcdir(os.path.join(self.common_test_dir, '40 options'))
        self.init(testdir)
        with open(os.path.join(testdir, 'meson_options.txt'), 'a', encoding='utf-8') as f:
            f.write("option('new_option', type : 'boolean', value : false)")
        self.setconf('-Dnew_option=true')
        self.assertEqual(self.getconf('new_option'), True)

    def test_configure_removed_option(self) -> None:
        """Removing an options without reconfiguring should still give an error."""
        testdir = self.copy_srcdir(os.path.join(self.common_test_dir, '40 options'))
        self.init(testdir)
        with open(os.path.join(testdir, 'meson_options.txt'), 'r', encoding='utf-8') as f:
            opts = f.readlines()
        with open(os.path.join(testdir, 'meson_options.txt'), 'w', encoding='utf-8') as f:
            for line in opts:
                if line.startswith("option('neg'"):
                    continue
                f.write(line)
        with self.assertRaises(subprocess.CalledProcessError) as e:
            self.setconf('-Dneg_int_opt=0')
        self.assertIn('Unknown option: "neg_int_opt"', e.exception.stdout)

    def test_reconfigure_option(self) -> None:
        testdir = self.copy_srcdir(os.path.join(self.common_test_dir, '40 options'))
        self.init(testdir)
        self.assertEqual(self.getconf('neg_int_opt'), -3)
        with self.assertRaises(subprocess.CalledProcessError) as e:
            self.init(testdir, extra_args=['--reconfigure', '-Dneg_int_opt=0'])
        self.assertEqual(self.getconf('neg_int_opt'), -3)
        self.init(testdir, extra_args=['--reconfigure', '-Dneg_int_opt=-2'])
        self.assertEqual(self.getconf('neg_int_opt'), -2)

    def test_configure_option_changed_constraints(self) -> None:
        """Changing the constraints of an option without reconfiguring should work."""
        testdir = self.copy_srcdir(os.path.join(self.common_test_dir, '40 options'))
        self.init(testdir)
        with open(os.path.join(testdir, 'meson_options.txt'), 'r', encoding='utf-8') as f:
            opts = f.readlines()
        with open(os.path.join(testdir, 'meson_options.txt'), 'w', encoding='utf-8') as f:
            for line in opts:
                if line.startswith("option('neg'"):
                    f.write("option('neg_int_opt', type : 'integer', min : -10, max : 10, value : -3)\n")
                else:
                    f.write(line)
        self.setconf('-Dneg_int_opt=-10')
        self.assertEqual(self.getconf('neg_int_opt'), -10)

    def test_configure_meson_options_txt_to_meson_options(self) -> None:
        """Changing from a meson_options.txt to meson.options should still be detected."""
        testdir = self.copy_srcdir(os.path.join(self.common_test_dir, '40 options'))
        self.init(testdir)
        with open(os.path.join(testdir, 'meson_options.txt'), 'r', encoding='utf-8') as f:
            opts = f.readlines()
        with open(os.path.join(testdir, 'meson_options.txt'), 'w', encoding='utf-8') as f:
            for line in opts:
                if line.startswith("option('neg'"):
                    f.write("option('neg_int_opt', type : 'integer', min : -10, max : 10, value : -3)\n")
                else:
                    f.write(line)
        shutil.move(os.path.join(testdir, 'meson_options.txt'), os.path.join(testdir, 'meson.options'))
        self.setconf('-Dneg_int_opt=-10')
        self.assertEqual(self.getconf('neg_int_opt'), -10)

    def test_configure_options_file_deleted(self) -> None:
        """Deleting all option files should make seting a project option an error."""
        testdir = self.copy_srcdir(os.path.join(self.common_test_dir, '40 options'))
        self.init(testdir)
        os.unlink(os.path.join(testdir, 'meson_options.txt'))
        with self.assertRaises(subprocess.CalledProcessError) as e:
            self.setconf('-Dneg_int_opt=0')
        self.assertIn('Unknown option: "neg_int_opt"', e.exception.stdout)

    def test_configure_options_file_added(self) -> None:
        """A new project option file should be detected."""
        testdir = self.copy_srcdir(os.path.join(self.common_test_dir, '1 trivial'))
        self.init(testdir)
        with open(os.path.join(testdir, 'meson.options'), 'w', encoding='utf-8') as f:
            f.write("option('new_option', type : 'string', value : 'foo')")
        self.setconf('-Dnew_option=bar')
        self.assertEqual(self.getconf('new_option'), 'bar')

    def test_configure_options_file_added_old(self) -> None:
        """A new project option file should be detected."""
        testdir = self.copy_srcdir(os.path.join(self.common_test_dir, '1 trivial'))
        self.init(testdir)
        with open(os.path.join(testdir, 'meson_options.txt'), 'w', encoding='utf-8') as f:
            f.write("option('new_option', type : 'string', value : 'foo')")
        self.setconf('-Dnew_option=bar')
        self.assertEqual(self.getconf('new_option'), 'bar')

    def test_configure_new_option_subproject(self) -> None:
        """Adding a new option to a subproject without reconfiguring should work."""
        testdir = self.copy_srcdir(os.path.join(self.common_test_dir, '43 subproject options'))
        self.init(testdir)
        with open(os.path.join(testdir, 'subprojects/subproject/meson_options.txt'), 'a', encoding='utf-8') as f:
            f.write("option('new_option', type : 'boolean', value : false)")
        self.setconf('-Dsubproject:new_option=true')
        self.assertEqual(self.getconf('subproject:new_option'), True)

    def test_mtest_rebuild_deps(self):
        testdir = os.path.join(self.unit_test_dir, '106 underspecified mtest')
        self.init(testdir)

        with self.assertRaises(subprocess.CalledProcessError):
            self._run(self.mtest_command)
        self.clean()

        with self.assertRaises(subprocess.CalledProcessError):
            self._run(self.mtest_command + ['runner-without-dep'])
        self.clean()

        self._run(self.mtest_command + ['runner-with-exedep'])

    def test_buildtype_debug_optimization_override(self) -> None:
        """Explicitly setting debug or optimization should override buildtype."""
        testdir = os.path.join(self.common_test_dir, '1 trivial')

        for extra_args in [['--debug', '--buildtype=release'],
                           ['-Ddebug=true', '--buildtype=release'],
                           ['-Ddebug=true', '-Dbuildtype=release'],
                           ['--debug', '-Dbuildtype=release']]:
            with self.subTest(extra_args=extra_args):
                self.new_builddir()
                self.init(testdir, extra_args=extra_args)
                opts = self.introspect('--buildoptions')
                self.assertEqual(self.getconf('buildtype', opts), 'release')
                self.assertEqual(self.getconf('debug', opts), True)
                self.assertEqual(self.getconf('optimization', opts), '3')

        for extra_args in [['--optimization=3', '--buildtype=debug'],
                           ['-Doptimization=3', '--buildtype=debug'],
                           ['-Doptimization=3', '-Dbuildtype=debug'],
                           ['--optimization=3', '-Dbuildtype=debug']]:
            with self.subTest(extra_args=extra_args):
                self.new_builddir()
                self.init(testdir, extra_args=extra_args)
                opts = self.introspect('--buildoptions')
                self.assertEqual(self.getconf('buildtype', opts), 'debug')
                self.assertEqual(self.getconf('debug', opts), True)
                self.assertEqual(self.getconf('optimization', opts), '3')

    def test_setup_mixed_long_short_options(self) -> None:
        """Mixing unity and unity_size as long and short options should work."""
        testdir = self.copy_srcdir(os.path.join(self.common_test_dir, '1 trivial'))
        self.init(testdir, extra_args=['-Dunity=on', '--unity-size=123'])

    def test_minit_executable_with_matching_source(self) -> None:
        """meson init --executable bar with bar.c present must not create a spurious project-name.c"""
        with tempfile.TemporaryDirectory() as d:
            Path(d, 'bar.c').write_text('int main() {}', encoding='utf-8')
            self._run(self.meson_command + ['init', '--name', 'foo', '--executable', 'bar', '--language', 'c'], workdir=d)
            files = {p.name for p in Path(d).iterdir()}
            self.assertIn('meson.build', files)
            self.assertNotIn('foo.c', files, 'spurious foo.c was created')
            meson_build = Path(d, 'meson.build').read_text(encoding='utf-8')
            self.assertNotIn("'foo.c'", meson_build, 'spurious foo.c listed in meson.build')
            self.assertIn("'bar.c'", meson_build, 'existing bar.c not listed in meson.build')

    def test_minit_multi_file_keeps_existing_sources(self) -> None:
        """meson init without --executable must create a new source file alongside existing ones"""
        with tempfile.TemporaryDirectory() as d:
            Path(d, 'bar.c').write_text('int bar() {}', encoding='utf-8')
            Path(d, 'foo.c').write_text('int main() {}', encoding='utf-8')
            self._run(self.meson_command + ['init', '--name', 'foo', '--language', 'c'], workdir=d)
            files = {p.name for p in Path(d).iterdir()}
            self.assertIn('meson.build', files)
            meson_build = Path(d, 'meson.build').read_text(encoding='utf-8')
            self.assertIn("'foo.c'", meson_build, 'existing foo.c not listed in meson.build')
            self.assertIn("'bar.c'", meson_build, 'existing bar.c not listed in meson.build')

    def test_minit_wrong_name_creates_sample(self) -> None:
        """meson init in an empty directory must create a sample source file"""
        with tempfile.TemporaryDirectory() as d:
            Path(d, 'bar.c').write_text('int main() {}', encoding='utf-8')
            self._run(self.meson_command + ['init', '--name', 'foo', '--language', 'c'], workdir=d)
            files = {p.name for p in Path(d).iterdir()}
            self.assertIn('meson.build', files)
            self.assertIn('foo.c', files)
            meson_build = Path(d, 'meson.build').read_text(encoding='utf-8')
            self.assertIn("'foo.c'", meson_build, 'created foo.c not listed in meson.build')
            self.assertNotIn("'bar.c'", meson_build, 'existing bar.c listed in meson.build')

    def test_minit_empty_dir_creates_sample(self) -> None:
        """meson init in an empty directory must create a sample source file"""
        with tempfile.TemporaryDirectory() as d:
            self._run(self.meson_command + ['init', '--name', 'foo', '--language', 'c'], workdir=d)
            files = {p.name for p in Path(d).iterdir()}
            self.assertIn('meson.build', files)
            self.assertIn('foo.c', files)
            meson_build = Path(d, 'meson.build').read_text(encoding='utf-8')
            self.assertIn("'foo.c'", meson_build, 'created foo.c not listed in meson.build')

    def test_readonly_sourcedir(self) -> None:
        """Test building with read-only source directory."""
        testdir = self.copy_srcdir(os.path.join(self.common_test_dir, '233 wrap case'))

        # Make the source directory and all its contents read-only recursively
        # Keep execute permission on directories
        for dir, _, files in os.walk(testdir):
            os.chmod(dir, 0o555)
            for file in files:
                filepath = os.path.join(dir, file)
                os.chmod(filepath, 0o444)

        self.init(testdir)
        self.build()

    @skip_if_not_language('java')
    def test_java_build_subdir_correct_deps(self):
        """Test that jar() with build_subdir sources does not rebuild unnecessarily."""
        testdir = os.path.join(self.src_root, 'test cases', 'java', '11 build subdir')
        self.init(testdir)
        self.build()
        self.assertBuildIsNoop()



# SPDX-License-Identifier: Apache-2.0
# Copyright 2016 The Meson development team

from __future__ import annotations

import argparse
import subprocess
import os
import typing as T

parser = argparse.ArgumentParser()
parser.add_argument('input')
parser.add_argument('output')
parser.add_argument('type')
parser.add_argument('podir')
parser.add_argument('--msgfmt', default='msgfmt')
parser.add_argument('--datadirs', default='')
parser.add_argument('args', default=[], metavar='extra msgfmt argument', nargs='*')


def run(args: T.List[str]) -> int:
    options = parser.parse_args(args)
    env = None
    if options.datadirs:
        env = os.environ.copy()
        env.update({'GETTEXTDATADIRS': options.datadirs})
    return subprocess.call([options.msgfmt, '--' + options.type, '-d', options.podir,
                            '--template', options.input,  '-o', options.output] + options.args,
                           env=env)
